#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Dict mapping custom schema instances from eln_data.yaml file on concepts in NXem."""

# pylint: disable=no-member,line-too-long

# releasing line-too-long restriction to avoid having line breaks in the mapping table
# made the experience that when having a widescreen working with the mapping table
# as single-line instructions is more convenient to read and parsable by human eye

# there are several issues with the current design of how data from the eln_data.yaml are passed
# some quantities from a custom schema instance end up as list of dictionaries, like here
# aperture and user, however in this mapping approach they would demand concept-specific
# modifiers to be picked up by, currently the em_spctrscpy, em_om, apm examples individually
# parse relevant quantities for each section which makes the code difficult to read and
# unnecessarily lengthy, with em_nion we would like to test if instead we can use
# a set of mapping tables whereby to read content from a custom schema instance eln_data.yaml
# results file directly into the template which the em_nion reader has to fill and pass
# then to the data converter

NxEmElnInput = {"IGNORE": {"fun": "load_from_dict_list", "terms": "em_lab/detector"},
                "IGNORE": {"fun": "load_from", "terms": "em_lab/ebeam_column/aberration_correction/applied"},
                "IGNORE": {"fun": "load_from_dict_list", "terms": "em_lab/ebeam_column/aperture_em"},
                "/ENTRY[entry*]/em_lab/EBEAM_COLUMN[ebeam_column]/electron_source/emitter_type": {"fun": "load_from", "terms": "em_lab/ebeam_column/electron_source/emitter_type"},
                "/ENTRY[entry*]/em_lab/EBEAM_COLUMN[ebeam_column]/electron_source/voltage/@units": {"fun": "load_from", "terms": "em_lab/ebeam_column/electron_source/voltage/unit"},
                "/ENTRY[entry*]/em_lab/EBEAM_COLUMN[ebeam_column]/electron_source/voltage": {"fun": "load_from", "terms": "em_lab/ebeam_column/electron_source/voltage/value"},
                "/ENTRY[entry*]/em_lab/FABRICATION[fabrication]/capabilities": {"fun": "load_from", "terms": "em_lab/fabrication/capabilities"},
                "/ENTRY[entry*]/em_lab/FABRICATION[fabrication]/identifier": {"fun": "load_from", "terms": "em_lab/fabrication/identifier"},
                "/ENTRY[entry*]/em_lab/FABRICATION[fabrication]/model": {"fun": "load_from", "terms": "em_lab/fabrication/model"},
                "/ENTRY[entry*]/em_lab/FABRICATION[fabrication]/vendor": {"fun": "load_from", "terms": "em_lab/fabrication/vendor"},
                "/ENTRY[entry*]/em_lab/instrument_name": {"fun": "load_from", "terms": "em_lab/instrument_name"},
                "/ENTRY[entry*]/em_lab/location": {"fun": "load_from", "terms": "em_lab/location"},
                "IGNORE": {"fun": "load_from", "terms": "em_lab/optical_system_em/beam_current/unit"},
                "IGNORE": {"fun": "load_from", "terms": "em_lab/optical_system_em/beam_current/value"},
                "IGNORE": {"fun": "load_from", "terms": "em_lab/optical_system_em/beam_current_description"},
                "IGNORE": {"fun": "load_from", "terms": "em_lab/optical_system_em/magnification"},
                "IGNORE": {"fun": "load_from", "terms": "em_lab/optical_system_em/semi_convergence_angle/unit"},
                "IGNORE": {"fun": "load_from", "terms": "em_lab/optical_system_em/semi_convergence_angle/value"},
                "/ENTRY[entry*]/em_lab/stage_lab/description": {"fun": "load_from", "terms": "em_lab/stage_lab/description"},
                "/ENTRY[entry*]/em_lab/stage_lab/name": {"fun": "load_from", "terms": "em_lab/stage_lab/name"},
                "/ENTRY[entry*]/@version": {"fun": "load_from", "terms": "entry/attr_version"},
                "/ENTRY[entry*]/definition": {"fun": "load_from", "terms": "entry/definition"},
                "/ENTRY[entry*]/end_time": {"fun": "load_from", "terms": "entry/end_time"},
                "/ENTRY[entry*]/experiment_description": {"fun": "load_from", "terms": "entry/experiment_description"},
                "/ENTRY[entry*]/experiment_identifier": {"fun": "load_from", "terms": "entry/experiment_identifier"},
                "/ENTRY[entry*]/PROGRAM[program*]/program": {"fun": "load_from", "terms": "entry/program"},
                "/ENTRY[entry*]/PROGRAM[program*]/program/@version": {"fun": "load_from", "terms": "entry/program__attr_version"},
                "/ENTRY[entry*]/start_time": {"fun": "load_from", "terms": "entry/start_time"},
                "IGNORE": {"fun": "load_from_list_of_dict", "terms": "user"}}

# NeXus concept specific mapping tables which require special treatment as the current
# NOMAD OASIS custom schema implementation delivers them as a list of dictionaries instead
# of a directly flattenable list of keyword, value pairs

NxApertureEmFromListOfDict = {"/ENTRY[entry*]/em_lab/EBEAM_COLUMN[ebeam_column]/APERTURE_EM[aperture_em*]/name": {"fun": "load_from", "terms": "name"},
                              "/ENTRY[entry*]/em_lab/EBEAM_COLUMN[ebeam_column]/APERTURE_EM[aperture_em*]/value": {"fun": "load_from", "terms": "value"}}

NxUserFromListOfDict = {"/ENTRY[entry*]/USER[user*]/name": {"fun": "load_from", "terms": "name"},
                        "/ENTRY[entry*]/USER[user*]/affiliation": {"fun": "load_from", "terms": "affiliation"},
                        "/ENTRY[entry*]/USER[user*]/address": {"fun": "load_from", "terms": "address"},
                        "/ENTRY[entry*]/USER[user*]/email": {"fun": "load_from", "terms": "email"},
                        "/ENTRY[entry*]/USER[user*]/orcid": {"fun": "load_from", "terms": "orcid"},
                        "/ENTRY[entry*]/USER[user*]/orcid_platform": {"fun": "load_from", "terms": "orcid_platform"},
                        "/ENTRY[entry*]/USER[user*]/telephone_number": {"fun": "load_from", "terms": "telephone_number"},
                        "/ENTRY[entry*]/USER[user*]/role": {"fun": "load_from", "terms": "role"},
                        "/ENTRY[entry*]/USER[user*]/social_media_name": {"fun": "load_from", "terms": "social_media_name"},
                        "/ENTRY[entry*]/USER[user*]/social_media_platform": {"fun": "load_from", "terms": "social_media_platform"}}

NxDetectorListOfDict = {"/ENTRY[entry*]/em_lab/DETECTOR[detector*]/local_name": {"fun": "load_from", "terms": "local_name"}}

# atom_types is a good example for specific cases where one cannot just blindly map
# the list that comes from the custom schema ELN instance, because
# people may enter invalid types of atoms (which would generate problems in NOMAD OASIS)
# and for NeXus we would like to have a "string of a comma-separated list of element names"

NxSample = {"IGNORE": {"fun": "load_from", "terms": "sample/atom_types"},
            "/ENTRY[entry*]/sample/description": {"fun": "load_from", "terms": "sample/description"},
            "/ENTRY[entry*]/sample/method": {"fun": "load_from", "terms": "sample/method"},
            "/ENTRY[entry*]/sample/name": {"fun": "load_from", "terms": "sample/name"},
            "/ENTRY[entry*]/sample/preparation_date": {"fun": "load_from", "terms": "sample/preparation_date"},
            "/ENTRY[entry*]/sample/sample_history": {"fun": "load_from", "terms": "sample/sample_history"},
            "/ENTRY[entry*]/sample/short_title": {"fun": "load_from", "terms": "sample/short_title"},
            "/ENTRY[entry*]/sample/thickness": {"fun": "load_from", "terms": "sample/thickness/value"},
            "/ENTRY[entry*]/sample/thickness/@units": {"fun": "load_from", "terms": "sample/thickness/unit"}}



# pylint: disable=too-many-lines
"""Parse NeXus definition files
"""

import os
import textwrap
import xml.etree.ElementTree as ET
from functools import lru_cache
from glob import glob


class NxdlAttributeError(Exception):
    """An exception for throwing an error when an Nxdl attribute is not found."""


def get_app_defs_names():
    """Returns all the AppDef names without their extension: .nxdl.xml"""
    app_def_path_glob = (
        f"{get_nexus_definitions_path()}{os.sep}applications{os.sep}*.nxdl*"
    )
    contrib_def_path_glob = (
        f"{get_nexus_definitions_path()}{os.sep}"
        f"contributed_definitions{os.sep}*.nxdl*"
    )
    files = sorted(glob(app_def_path_glob)) + sorted(glob(contrib_def_path_glob))
    return [os.path.basename(file).split(".")[0] for file in files] + ["NXroot"]


@lru_cache(maxsize=None)
def get_xml_root(file_path):
    """Reducing I/O time by caching technique"""

    return ET.parse(file_path).getroot()


def get_nexus_definitions_path():
    """Check NEXUS_DEF_PATH variable.
    If it is empty, this function is filling it"""
    try:  # either given by sys env
        return os.environ["NEXUS_DEF_PATH"]
    except KeyError:  # or it should be available locally under the dir 'definitions'
        local_dir = os.path.abspath(os.path.dirname(__file__))
        return os.path.join(local_dir, f"..{os.sep}..")


def get_hdf_root(hdf_node):
    """Get the root HDF5 node"""
    node = hdf_node
    while node.name != "/":
        node = node.parent
    return node


def get_hdf_parent(hdf_info):
    """Get the parent of an hdf_node in an hdf_info"""
    if "hdf_path" not in hdf_info:
        return hdf_info["hdf_node"].parent
    node = (
        get_hdf_root(hdf_info["hdf_node"])
        if "hdf_root" not in hdf_info
        else hdf_info["hdf_root"]
    )
    for child_name in hdf_info["hdf_path"].split("/"):
        node = node[child_name]
    return node


def get_parent_path(hdf_name):
    """Get parent path"""
    return "/".join(hdf_name.split("/")[:-1])


def get_hdf_info_parent(hdf_info):
    """Get the hdf_info for the parent of an hdf_node in an hdf_info"""
    if "hdf_path" not in hdf_info:
        return {"hdf_node": hdf_info["hdf_node"].parent}
    node = (
        get_hdf_root(hdf_info["hdf_node"])
        if "hdf_root" not in hdf_info
        else hdf_info["hdf_root"]
    )
    for child_name in hdf_info["hdf_path"].split("/")[1:-1]:
        node = node[child_name]
    return {"hdf_node": node, "hdf_path": get_parent_path(hdf_info["hdf_path"])}


def get_nx_class(nxdl_elem):
    """Get the nexus class for a NXDL node"""
    if "category" in nxdl_elem.attrib.keys():
        return None
    try:
        return nxdl_elem.attrib["type"]
    except KeyError:
        return "NX_CHAR"


def get_nx_namefit(hdf_name, name, name_any=False):
    """Checks if an HDF5 node name corresponds to a child of the NXDL element
    uppercase letters in front can be replaced by arbitraty name, but
    uppercase to lowercase match is preferred,
    so such match is counted as a measure of the fit"""
    if name == hdf_name:
        return len(name) * 2
    # count leading capitals
    counting = 0
    while counting < len(name) and name[counting].upper() == name[counting]:
        counting += 1
    if (
        name_any
        or counting == len(name)
        or (counting > 0 and hdf_name.endswith(name[counting:]))
    ):  # if potential fit
        # count the matching chars
        fit = 0
        for i in range(min(counting, len(hdf_name))):
            if hdf_name[i].upper() == name[i]:
                fit += 1
            else:
                break
        if fit == min(counting, len(hdf_name)):  # accept only full fits as better fits
            return fit
        return 0
    return -1  # no fit


def get_nx_classes():
    """Read base classes from the NeXus definition folder.
    Check each file in base_classes, applications, contributed_definitions.
    If its category attribute is 'base', then it is added to the list."""
    base_classes = sorted(
        glob(os.path.join(get_nexus_definitions_path(), "base_classes", "*.nxdl.xml"))
    )
    applications = sorted(
        glob(os.path.join(get_nexus_definitions_path(), "applications", "*.nxdl.xml"))
    )
    contributed = sorted(
        glob(
            os.path.join(
                get_nexus_definitions_path(), "contributed_definitions", "*.nxdl.xml"
            )
        )
    )
    nx_clss = []
    for nexus_file in base_classes + applications + contributed:
        root = get_xml_root(nexus_file)
        if root.attrib["category"] == "base":
            nx_clss.append(str(nexus_file[nexus_file.rindex(os.sep) + 1 :])[:-9])
    nx_clss = sorted(nx_clss)
    return nx_clss


def get_nx_units():
    """Read unit kinds from the NeXus definition/nxdlTypes.xsd file"""
    filepath = f"{get_nexus_definitions_path()}{os.sep}nxdlTypes.xsd"
    root = get_xml_root(filepath)
    units_and_type_list = []
    for child in root:
        for i in child.attrib.values():
            units_and_type_list.append(i)
    flag = False
    for line in units_and_type_list:
        if line == "anyUnitsAttr":
            flag = True
            nx_units = []
        elif "NX" in line and flag is True:
            nx_units.append(line)
        elif line == "primitiveType":
            flag = False
        else:
            pass
    return nx_units


def get_nx_attribute_type():
    """Read attribute types from the NeXus definition/nxdlTypes.xsd file"""
    filepath = get_nexus_definitions_path() + "/nxdlTypes.xsd"
    root = get_xml_root(filepath)
    units_and_type_list = []
    for child in root:
        for i in child.attrib.values():
            units_and_type_list.append(i)
    flag = False
    for line in units_and_type_list:
        if line == "primitiveType":
            flag = True
            nx_types = []
        elif "NX" in line and flag is True:
            nx_types.append(line)
        elif line == "anyUnitsAttr":
            flag = False
        else:
            pass
    return nx_types


def get_node_name(node):
    """Node - xml node. Returns html documentation name.
    Either as specified by the 'name' or taken from the type (nx_class).
    Note that if only class name is available, the NX prefix is removed and
    the string is converted to UPPER case."""
    if "name" in node.attrib.keys():
        name = node.attrib["name"]
    else:
        name = node.attrib["type"]
        if name.startswith("NX"):
            name = name[2:].upper()
    return name


def belongs_to(nxdl_elem, child, name, class_type=None, hdf_name=None):
    """Checks if an HDF5 node name corresponds to a child of the NXDL element
    uppercase letters in front can be replaced by arbitraty name, but
    uppercase to lowercase match is preferred"""
    if class_type and get_nx_class(child) != class_type:
        return False
    act_htmlname = get_node_name(child)
    chk_name = hdf_name or name
    if act_htmlname == chk_name:
        return True
    if not hdf_name:  # search for name fits is only allowed for hdf_nodes
        return False
    try:  # check if nameType allows different name
        name_any = bool(child.attrib["nameType"] == "any")
    except KeyError:
        name_any = False
    params = [act_htmlname, chk_name, name_any, nxdl_elem, child, name]
    return belongs_to_capital(params)


def belongs_to_capital(params):
    """Checking continues for Upper case"""
    (act_htmlname, chk_name, name_any, nxdl_elem, child, name) = params
    # or starts with capital and no reserved words used
    if (
        (name_any or "A" <= act_htmlname[0] <= "Z")
        and name != "doc"
        and name != "enumeration"
    ):
        fit = get_nx_namefit(chk_name, act_htmlname, name_any)  # check if name fits
        if fit < 0:
            return False
        for child2 in nxdl_elem:
            if (
                get_local_name_from_xml(child) != get_local_name_from_xml(child2)
                or get_node_name(child2) == act_htmlname
            ):
                continue
            # check if the name of another sibling fits better
            name_any2 = (
                "nameType" in child2.attrib.keys()
                and child2.attrib["nameType"] == "any"
            )
            fit2 = get_nx_namefit(chk_name, get_node_name(child2), name_any2)
            if fit2 > fit:
                return False
        # accept this fit
        return True
    return False


def get_local_name_from_xml(element):
    """Helper function to extract the element tag without the namespace."""
    return element.tag[element.tag.rindex("}") + 1 :]


def get_own_nxdl_child_reserved_elements(child, name, nxdl_elem):
    """checking reserved elements, like doc, enumeration"""
    if get_local_name_from_xml(child) == "doc" and name == "doc":
        if nxdl_elem.get("nxdlbase"):
            child.set("nxdlbase", nxdl_elem.get("nxdlbase"))
            child.set("nxdlbase_class", nxdl_elem.get("nxdlbase_class"))
            child.set("nxdlpath", nxdl_elem.get("nxdlpath") + "/doc")
        return child
    if get_local_name_from_xml(child) == "enumeration" and name == "enumeration":
        if nxdl_elem.get("nxdlbase"):
            child.set("nxdlbase", nxdl_elem.get("nxdlbase"))
            child.set("nxdlbase_class", nxdl_elem.get("nxdlbase_class"))
            child.set("nxdlpath", nxdl_elem.get("nxdlpath") + "/enumeration")
        return child
    return False


def get_own_nxdl_child_base_types(child, class_type, nxdl_elem, name, hdf_name):
    """checking base types of group, field,m attribute"""
    if get_local_name_from_xml(child) == "group":
        if (
            class_type is None or (class_type and get_nx_class(child) == class_type)
        ) and belongs_to(nxdl_elem, child, name, class_type, hdf_name):
            if nxdl_elem.get("nxdlbase"):
                child.set("nxdlbase", nxdl_elem.get("nxdlbase"))
                child.set("nxdlbase_class", nxdl_elem.get("nxdlbase_class"))
                child.set(
                    "nxdlpath", nxdl_elem.get("nxdlpath") + "/" + get_node_name(child)
                )
            return child
    if get_local_name_from_xml(child) == "field" and belongs_to(
        nxdl_elem, child, name, None, hdf_name
    ):
        if nxdl_elem.get("nxdlbase"):
            child.set("nxdlbase", nxdl_elem.get("nxdlbase"))
            child.set("nxdlbase_class", nxdl_elem.get("nxdlbase_class"))
            child.set(
                "nxdlpath", nxdl_elem.get("nxdlpath") + "/" + get_node_name(child)
            )
        return child
    if get_local_name_from_xml(child) == "attribute" and belongs_to(
        nxdl_elem, child, name, None, hdf_name
    ):
        if nxdl_elem.get("nxdlbase"):
            child.set("nxdlbase", nxdl_elem.get("nxdlbase"))
            child.set("nxdlbase_class", nxdl_elem.get("nxdlbase_class"))
            child.set(
                "nxdlpath", nxdl_elem.get("nxdlpath") + "/" + get_node_name(child)
            )
        return child
    return False


def get_own_nxdl_child(
    nxdl_elem, name, class_type=None, hdf_name=None, nexus_type=None
):
    """Checks if an NXDL child node fits to the specific name (either nxdl or hdf)
    name       - nxdl name
    class_type - nxdl type or hdf classname (for groups, it is obligatory)
    hdf_name   - hdf name"""
    for child in nxdl_elem:
        if "name" in child.attrib and child.attrib["name"] == name:
            if nxdl_elem.get("nxdlbase"):
                child.set("nxdlbase", nxdl_elem.get("nxdlbase"))
                child.set("nxdlbase_class", nxdl_elem.get("nxdlbase_class"))
                child.set(
                    "nxdlpath", nxdl_elem.get("nxdlpath") + "/" + get_node_name(child)
                )
            return child
    for child in nxdl_elem:
        if "name" in child.attrib and child.attrib["name"] == name:
            child.set("nxdlbase", nxdl_elem.get("nxdlbase"))
            return child

    for child in nxdl_elem:
        result = get_own_nxdl_child_reserved_elements(child, name, nxdl_elem)
        if result is not False:
            return result
        if nexus_type and get_local_name_from_xml(child) != nexus_type:
            continue
        result = get_own_nxdl_child_base_types(
            child, class_type, nxdl_elem, name, hdf_name
        )
        if result is not False:
            return result
    return None


def find_definition_file(bc_name):
    """find the nxdl file corresponding to the name.
    Note that it first checks in contributed and goes beyond only if no contributed found
    """
    bc_filename = None
    for nxdl_folder in ["contributed_definitions", "base_classes", "applications"]:
        if os.path.exists(
            f"{get_nexus_definitions_path()}{os.sep}"
            f"{nxdl_folder}{os.sep}{bc_name}.nxdl.xml"
        ):
            bc_filename = (
                f"{get_nexus_definitions_path()}{os.sep}"
                f"{nxdl_folder}{os.sep}{bc_name}.nxdl.xml"
            )
            break
    return bc_filename


def get_nxdl_child(
    nxdl_elem, name, class_type=None, hdf_name=None, nexus_type=None, go_base=True
):  # pylint: disable=too-many-arguments
    """Get the NXDL child node corresponding to a specific name
    (e.g. of an HDF5 node,or of a documentation) note that if child is not found in application
    definition, it also checks for the base classes"""
    # search for possible fits for hdf_nodes : skipped
    # only exact hits are returned when searching an nxdl child
    own_child = get_own_nxdl_child(nxdl_elem, name, class_type, hdf_name, nexus_type)
    if own_child is not None:
        return own_child
    if not go_base:
        return None
    bc_name = get_nx_class(nxdl_elem)  # check in the base class, app def or contributed
    if bc_name[2] == "_":  # filter primitive types
        return None
    if (
        bc_name == "group"
    ):  # Check if it is the root element. Then send to NXroot.nxdl.xml
        bc_name = "NXroot"
    bc_filename = find_definition_file(bc_name)
    if not bc_filename:
        raise ValueError("nxdl file not found in definitions folder!")
    bc_obj = ET.parse(bc_filename).getroot()
    bc_obj.set("nxdlbase", bc_filename)
    if "category" in bc_obj.attrib:
        bc_obj.set("nxdlbase_class", bc_obj.attrib["category"])
    bc_obj.set("nxdlpath", "")
    return get_own_nxdl_child(bc_obj, name, class_type, hdf_name, nexus_type)


def get_required_string(nxdl_elem):
    """Check for being REQUIRED, RECOMMENDED, OPTIONAL, NOT IN SCHEMA"""
    if nxdl_elem is None:
        return "<<NOT IN SCHEMA>>"
    is_optional = (
        "optional" in nxdl_elem.attrib.keys() and nxdl_elem.attrib["optional"] == "true"
    )
    is_minoccurs = (
        "minOccurs" in nxdl_elem.attrib.keys() and nxdl_elem.attrib["minOccurs"] == "0"
    )
    is_recommended = (
        "recommended" in nxdl_elem.attrib.keys()
        and nxdl_elem.attrib["recommended"] == "true"
    )

    if is_recommended:
        return "<<RECOMMENDED>>"
    if is_optional or is_minoccurs:
        return "<<OPTIONAL>>"
    # default optionality: in BASE CLASSES is true; in APPLICATIONS is false
    try:
        if nxdl_elem.get("nxdlbase_class") == "base":
            return "<<OPTIONAL>>"
    except TypeError:
        return "<<REQUIRED>>"
    return "<<REQUIRED>>"


# below there are some functions used in get_nxdl_doc function:
def write_doc_string(logger, doc, attr):
    """Simple function that prints a line in the logger if doc exists"""
    if doc:
        logger.debug("@" + attr + " [NX_CHAR]")
    return logger, doc, attr


def try_find_units(logger, elem, nxdl_path, doc, attr):
    """Try to find if units is defined inside the field in the NXDL element,
    otherwise try to find if units is defined as a child of the NXDL element."""
    try:  # try to find if units is defined inside the field in the NXDL element
        unit = elem.attrib[attr]
        if doc:
            logger.debug(get_node_concept_path(elem) + "@" + attr + " [" + unit + "]")
        elem = None
        nxdl_path.append(attr)
    except (
        KeyError
    ):  # otherwise try to find if units is defined as a child of the NXDL element
        orig_elem = elem
        elem = get_nxdl_child(elem, attr, nexus_type="attribute")
        if elem is not None:
            if doc:
                logger.debug(
                    get_node_concept_path(orig_elem)
                    + "@"
                    + attr
                    + " - ["
                    + get_nx_class(elem)
                    + "]"
                )
            nxdl_path.append(elem)
        else:  # if no units category were defined in NXDL:
            if doc:
                logger.debug(
                    get_node_concept_path(orig_elem)
                    + "@"
                    + attr
                    + " - REQUIRED, but undefined unit category"
                )
            nxdl_path.append(attr)
    return logger, elem, nxdl_path, doc, attr


def check_attr_name_nxdl(param):
    """Check for ATTRIBUTENAME_units in NXDL (normal).
    If not defined, check for ATTRIBUTENAME to see if the ATTRIBUTE
    is in the SCHEMA, but no units category were defined."""
    (logger, elem, nxdl_path, doc, attr, req_str) = param
    orig_elem = elem
    elem2 = get_nxdl_child(elem, attr, nexus_type="attribute")
    if elem2 is not None:  # check for ATTRIBUTENAME_units in NXDL (normal)
        elem = elem2
        if doc:
            logger.debug(
                get_node_concept_path(orig_elem)
                + "@"
                + attr
                + " - ["
                + get_nx_class(elem)
                + "]"
            )
        nxdl_path.append(elem)
    else:
        # if not defined, check for ATTRIBUTENAME to see if the ATTRIBUTE
        # is in the SCHEMA, but no units category were defined
        elem2 = get_nxdl_child(elem, attr[:-6], nexus_type="attribute")
        if elem2 is not None:
            req_str = "<<RECOMMENDED>>"
            if doc:
                logger.debug(
                    get_node_concept_path(orig_elem)
                    + "@"
                    + attr
                    + " - RECOMMENDED, but undefined unit category"
                )
            nxdl_path.append(attr)
        else:  # otherwise: NOT IN SCHEMA
            elem = elem2
            if doc:
                logger.debug(
                    get_node_concept_path(orig_elem)
                    + "@"
                    + attr
                    + " - IS NOT IN SCHEMA"
                )
    return logger, elem, nxdl_path, doc, attr, req_str


def try_find_default(
    logger, orig_elem, elem, nxdl_path, doc, attr
):  # pylint: disable=too-many-arguments
    """Try to find if default is defined as a child of the NXDL element"""
    if elem is not None:
        if doc:
            logger.debug(
                get_node_concept_path(orig_elem)
                + "@"
                + attr
                + " - ["
                + get_nx_class(elem)
                + "]"
            )
        nxdl_path.append(elem)
    else:  # if no default category were defined in NXDL:
        if doc:
            logger.debug(get_node_concept_path(orig_elem) + "@" + attr + " - [NX_CHAR]")
        nxdl_path.append(attr)
    return logger, elem, nxdl_path, doc, attr


def other_attrs(
    logger, orig_elem, elem, nxdl_path, doc, attr
):  # pylint: disable=too-many-arguments
    """Handle remaining attributes"""
    if elem is not None:
        if doc:
            logger.debug(
                get_node_concept_path(orig_elem)
                + "@"
                + attr
                + " - ["
                + get_nx_class(elem)
                + "]"
            )
        nxdl_path.append(elem)
    else:
        if doc:
            logger.debug(
                get_node_concept_path(orig_elem) + "@" + attr + " - IS NOT IN SCHEMA"
            )
    return logger, elem, nxdl_path, doc, attr


def get_node_concept_path(elem):
    """get the short version of nxdlbase:nxdlpath"""
    return str(elem.get("nxdlbase").split("/")[-1] + ":" + elem.get("nxdlpath"))


def get_doc(node, ntype, nxhtml, nxpath):
    """Get documentation"""
    # URL for html documentation
    anchor = ""
    for n_item in nxpath:
        anchor += n_item.lower() + "-"
    anchor = (
        "https://manual.nexusformat.org/classes/",
        nxhtml + "#" + anchor.replace("_", "-") + ntype,
    )
    if not ntype:
        anchor = anchor[:-1]
    doc = ""  # RST documentation from the field 'doc'
    doc_field = node.find("doc")
    if doc_field is not None:
        doc = doc_field.text
    (index, enums) = get_enums(node)  # enums
    if index:
        enum_str = (
            "\n "
            + ("Possible values:" if len(enums.split(",")) > 1 else "Obligatory value:")
            + "\n   "
            + enums
            + "\n"
        )
    else:
        enum_str = ""
    return anchor, doc + enum_str


def print_doc(node, ntype, level, nxhtml, nxpath):
    """Print documentation"""
    anchor, doc = get_doc(node, ntype, nxhtml, nxpath)
    print("  " * (level + 1) + anchor)
    preferred_width = 80 + level * 2
    wrapper = textwrap.TextWrapper(
        initial_indent="  " * (level + 1),
        width=preferred_width,
        subsequent_indent="  " * (level + 1),
        expand_tabs=False,
        tabsize=0,
    )
    if doc is not None:
        for par in doc.split("\n"):
            print(wrapper.fill(par))


def get_namespace(element):
    """Extracts the namespace for elements in the NXDL"""
    return element.tag[element.tag.index("{") : element.tag.rindex("}") + 1]


def get_enums(node):
    """Makes list of enumerations, if node contains any.
    Returns comma separated STRING of enumeration values, if there are enum tag,
    otherwise empty string."""
    # collect item values from enumeration tag, if any
    namespace = get_namespace(node)
    enums = []
    for enumeration in node.findall(f"{namespace}enumeration"):
        for item in enumeration.findall(f"{namespace}item"):
            enums.append(item.attrib["value"])
        enums = ",".join(enums)
        if enums != "":
            return (True, "[" + enums + "]")
    return (False, "")  # if there is no enumeration tag, returns empty string


def add_base_classes(elist, nx_name=None, elem: ET.Element = None):
    """Add the base classes corresponding to the last eleme in elist to the list. Note that if
    elist is empty, a nxdl file with the name of nx_name or a rather room elem is used if provided
    """
    if elist and nx_name is None:
        nx_name = get_nx_class(elist[-1])
    # to support recursive defintions, like NXsample in NXsample, the following test is removed
    # if elist and nx_name and f"{nx_name}.nxdl.xml" in (e.get('nxdlbase') for e in elist):
    #     return
    if elem is None:
        if not nx_name:
            return
        nxdl_file_path = find_definition_file(nx_name)
        if nxdl_file_path is None:
            nxdl_file_path = f"{nx_name}.nxdl.xml"
        elem = ET.parse(nxdl_file_path).getroot()
        elem.set("nxdlbase", nxdl_file_path)
    else:
        elem.set("nxdlbase", "")
    if "category" in elem.attrib:
        elem.set("nxdlbase_class", elem.attrib["category"])
    elem.set("nxdlpath", "")
    elist.append(elem)
    # add inherited base class
    if "extends" in elem.attrib and elem.attrib["extends"] != "NXobject":
        add_base_classes(elist, elem.attrib["extends"])
    else:
        add_base_classes(elist)


def set_nxdlpath(child, nxdl_elem):
    """
    Setting up child nxdlbase, nxdlpath and nxdlbase_class from nxdl_element.
    """
    if nxdl_elem.get("nxdlbase"):
        child.set("nxdlbase", nxdl_elem.get("nxdlbase"))
        child.set("nxdlbase_class", nxdl_elem.get("nxdlbase_class"))
        child.set("nxdlpath", nxdl_elem.get("nxdlpath") + "/" + get_node_name(child))
    return child


def get_direct_child(nxdl_elem, html_name):
    """returns the child of nxdl_elem which has a name
    corresponding to the the html documentation name html_name"""
    for child in nxdl_elem:
        if get_local_name_from_xml(child) in (
            "group",
            "field",
            "attribute",
        ) and html_name == get_node_name(child):
            decorated_child = set_nxdlpath(child, nxdl_elem)
            return decorated_child
    return None


def get_field_child(nxdl_elem, html_name):
    """returns the child of nxdl_elem which has a name
    corresponding to the html documentation name html_name"""
    data_child = None
    for child in nxdl_elem:
        if get_local_name_from_xml(child) != "field":
            continue
        if get_node_name(child) == html_name:
            data_child = set_nxdlpath(child, nxdl_elem)
            break
    return data_child


def get_best_nxdata_child(nxdl_elem, hdf_node, hdf_name):
    """returns the child of an NXdata nxdl_elem which has a name
    corresponding to the hdf_name"""
    nxdata = hdf_node.parent
    signals = []
    if "signal" in nxdata.attrs.keys():
        signals.append(nxdata.attrs.get("signal"))
    if "auxiliary_signals" in nxdata.attrs.keys():
        for aux_signal in nxdata.attrs.get("auxiliary_signals"):
            signals.append(aux_signal)
    data_child = get_field_child(nxdl_elem, "DATA")
    data_error_child = get_field_child(nxdl_elem, "FIELDNAME_errors")
    for signal in signals:
        if signal == hdf_name:
            return (data_child, 100)
        if hdf_name.endswith("_errors") and signal == hdf_name[:-7]:
            return (data_error_child, 100)
    axes = []
    if "axes" in nxdata.attrs.keys():
        for axis in nxdata.attrs.get("axes"):
            axes.append(axis)
    axis_child = get_field_child(nxdl_elem, "AXISNAME")
    for axis in axes:
        if axis == hdf_name:
            return (axis_child, 100)
    return (None, 0)


def get_best_child(nxdl_elem, hdf_node, hdf_name, hdf_class_name, nexus_type):
    """returns the child of nxdl_elem which has a name
    corresponding to the the html documentation name html_name"""
    bestfit = -1
    bestchild = None
    if (
        "name" in nxdl_elem.attrib.keys()
        and nxdl_elem.attrib["name"] == "NXdata"
        and hdf_node is not None
        and hdf_node.parent is not None
        and hdf_node.parent.attrs.get("NX_class") == "NXdata"
    ):
        (fnd_child, fit) = get_best_nxdata_child(nxdl_elem, hdf_node, hdf_name)
        if fnd_child is not None:
            return (fnd_child, fit)
    for child in nxdl_elem:
        fit = -2
        if get_local_name_from_xml(child) == nexus_type and (
            nexus_type != "group" or get_nx_class(child) == hdf_class_name
        ):
            name_any = (
                "nameType" in nxdl_elem.attrib.keys()
                and nxdl_elem.attrib["nameType"] == "any"
            )
            fit = get_nx_namefit(hdf_name, get_node_name(child), name_any)
        if fit > bestfit:
            bestfit = fit
            bestchild = set_nxdlpath(child, nxdl_elem)
    return (bestchild, bestfit)


def walk_elist(elist, html_name):
    """Handle elist from low priority inheritance classes to higher"""
    for ind in range(len(elist) - 1, -1, -1):
        child = get_direct_child(elist[ind], html_name)
        if child is None:
            # check for names fitting to a superclas definition
            main_child = None
            for potential_direct_parent in elist:
                main_child = get_direct_child(potential_direct_parent, html_name)
                if main_child is not None:
                    (fitting_child, _) = get_best_child(
                        elist[ind],
                        None,
                        html_name,
                        get_nx_class(main_child),
                        get_local_name_from_xml(main_child),
                    )
                    if fitting_child is not None:
                        child = fitting_child
                    break
        elist[ind] = child
        if elist[ind] is None:
            del elist[ind]
            continue
        # override: remove low priority inheritance classes if class_type is overriden
        if len(elist) > ind + 1 and get_nx_class(elist[ind]) != get_nx_class(
            elist[ind + 1]
        ):
            del elist[ind + 1 :]
        # add new base class(es) if new element brings such (and not a primitive type)
        if len(elist) == ind + 1 and get_nx_class(elist[ind])[0:3] != "NX_":
            add_base_classes(elist)
    return elist, html_name


@lru_cache(maxsize=None)
def get_inherited_nodes(
    nxdl_path: str = None,  # pylint: disable=too-many-arguments,too-many-locals
    nx_name: str = None,
    elem: ET.Element = None,
    attr=False,
):
    """Returns a list of ET.Element for the given path."""
    # let us start with the given definition file
    elist = []  # type: ignore[var-annotated]
    add_base_classes(elist, nx_name, elem)
    nxdl_elem_path = [elist[0]]

    class_path = []  # type: ignore[var-annotated]
    html_path = nxdl_path.split("/")[1:]
    path = html_path
    for pind in range(len(path)):
        html_name = html_path[pind]
        elist, html_name = walk_elist(elist, html_name)
        if elist:
            class_path.append(get_nx_class(elist[0]))
            nxdl_elem_path.append(elist[0])
    return (class_path, nxdl_elem_path, elist)


def get_node_at_nxdl_path(
    nxdl_path: str = None,
    nx_name: str = None,
    elem: ET.Element = None,
    exc: bool = True,
):
    """Returns an ET.Element for the given path.
    This function either takes the name for the NeXus Application Definition
    we are looking for or the root elem from a previously loaded NXDL file
    and finds the corresponding XML element with the needed attributes."""
    try:
        (class_path, nxdlpath, elist) = get_inherited_nodes(nxdl_path, nx_name, elem)
    except ValueError as value_error:
        if exc:
            raise NxdlAttributeError(
                f"Attributes were not found for {nxdl_path}. "
                "Please check this entry in the template dictionary."
            ) from value_error
        return None
    if class_path and nxdlpath and elist:
        elem = elist[0]
    else:
        elem = None
        if exc:
            raise NxdlAttributeError(
                f"Attributes were not found for {nxdl_path}. "
                "Please check this entry in the template dictionary."
            )
    return elem



#!/usr/bin/env python
"""Reads NeXus HDF5 files using nexusformat and prints the contents"""

from nexusformat.nexus import nxopen

fileName = "simple_example_basic.nexus.hdf5"
with nxopen(fileName) as f:
    print(f.tree)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Parse human-readable composition infos from set of ELN string text fields."""

# pylint: disable=no-member,too-many-branches

import re

import numpy as np

from ase.data import chemical_symbols


def parse_human_readable_composition_case_one(symbol):
    """Handle specification of matrix or remainder element."""
    return ("define_matrix", symbol, None, None, None)


def parse_human_readable_composition_case_two(args, symbol):
    """Handle case element and at.-% composition, no comp. stdev."""
    if args[1] in ["rem", "remainder", "matrix"]:
        return ("define_matrix", symbol, None, None, None)
    composition = re.match(r"[-+]?(?:\d*\.*\d+)", args[1])
    if composition is not None:
        fraction = np.float64(composition[0])
        return ("add_element", symbol, fraction, None, "at.-%")
    return (None, None, None, None, None)


def parse_human_readable_composition_case_three(human_input, args, symbol):
    """Handle case element with different than default normalization, no comp. stdev."""
    composition = re.findall(r"[-+]?(?:\d*\.*\d+)", human_input)
    if len(composition) == 1:
        fraction = np.float64(composition[0])
        normalization = args[2]
        if normalization in ["%", "at%", "at-%", "at.-%"]:
            return ("add_element", symbol, fraction, None, "at.-%")
        if normalization in ["wt%", "wt-%", "wt.-%"]:
            return ("add_element", symbol, fraction, None, "wt.-%")
        if normalization == "ppm":
            return ("add_element", symbol, fraction / 1.0e4, None, "at.-%")
        if normalization == "ppb":
            return ("add_element", symbol, fraction / 1.0e7, None, "at.-%")
    return (None, None, None, None, None)


def parse_human_readable_composition_case_four(human_input, symbol):
    """Handle case at.-% normalization with comp. stdev."""
    composition = re.findall(r"[-+]?(?:\d*\.*\d+)", human_input)
    composition_error = human_input.count("+-")
    if (len(composition) == 2) and (composition_error == 1):
        fraction = np.float64(composition[0])
        error = np.float64(composition[1])
        return ("add_element", symbol, fraction, error, "at.-%")
    return (None, None, None, None, None)


def parse_human_readable_composition_case_five(human_input, args, symbol):
    """Handle case with different than standard normalization and comp. stdev."""
    composition = re.findall(r"[-+]?(?:\d*\.*\d+)", human_input)
    if (len(composition) == 2) and (human_input.count("+-") == 1):
        fraction = np.float64(composition[0])
        error = np.float64(composition[1])
        normalization = args[2]
        if normalization in ["%", "at%", "at-%", "at.-%"]:
            return ("add_element", symbol, fraction, error, "at.-%")
        if normalization in ["wt%", "wt-%", "wt.-%"]:
            return ("add_element", symbol, fraction, error, "wt.-%")
        if normalization == "ppm":
            return ("add_element", symbol, fraction / 1.0e4, error / 1.0e4, "at.-%")
        if normalization == "ppb":
            return ("add_element", symbol, fraction / 1.0e7, error / 1.0e7, "at.-%")
    return (None, None, None, None, None)


def parse_human_readable_composition_information(eln_input):
    """Identify instruction to parse from eln_input to define composition table."""
    args = eln_input.split(" ")
    if len(args) >= 1:
        element_symbol = args[0]
        # composition value argument fraction is always expected in percent
        # i.e. human should have written 98 instead 0.98!
        if (element_symbol != "X") and (element_symbol in chemical_symbols):
            # case: "Mo"
            if len(args) == 1:
                return parse_human_readable_composition_case_one(
                    element_symbol)
            # case: "Mo matrix" or "Mo 98.0", always assuming at.-%!
            if len(args) == 2:
                return parse_human_readable_composition_case_two(
                    args, element_symbol)
            # case: "Mo 98 wt.-%", selectable at.-%, ppm, ppb, or wt.-%!
            if len(args) == 3:
                return parse_human_readable_composition_case_three(
                    eln_input, args, element_symbol)
            # case: "Mo 98 +- 2", always assuming at.-%!
            if len(args) == 4:
                return parse_human_readable_composition_case_four(
                    eln_input, element_symbol)
            # case: "Mo 98 wt.-% +- 2", selectable at.-%, ppm, ppb, or wt.-%!
            if len(args) == 5:
                return parse_human_readable_composition_case_five(
                    eln_input, args, element_symbol)
    return (None, None, None, None, None)


def parse_composition_table(composition_list):
    """Check if all the entries in the composition list yield a valid composition table."""
    composition_table = {}
    # check that there are no contradictions or inconsistenc
    for entry in composition_list:
        instruction, element, composition, stdev, normalization \
            = parse_human_readable_composition_information(entry)
        # print(f"{instruction}, {element}, {composition}, {stdev}, {normalization}")

        if instruction == "add_element":
            if "normalization" not in composition_table:
                if normalization is not None:
                    composition_table["normalization"] = normalization
            else:
                # as the normalization model is already defined, all following statements
                # need to comply because we assume we are not allowed to mix atom and weight
                # percent normalization in a composition_table
                if normalization is not None:
                    if normalization != composition_table["normalization"]:
                        raise ValueError("Composition list is contradicting as it \
                                         mixes atom- with weight-percent normalization!")

            if element not in composition_table:
                composition_table[element] = (composition, stdev)
            else:
                raise ValueError("Composition list is incorrectly formatted as if has \
                                 at least multiple lines for the same element!")
            continue
        if instruction == "define_matrix":
            if element not in composition_table:
                composition_table[element] = (None, None)
                # because the fraction is unclear at this point
            else:
                raise ValueError("Composition list is contradicting as it includes \
                                 at least two statements what the matrix should be!")

    # determine remaining fraction
    total_fractions = 0.
    remainder_element = None
    for keyword, tpl in composition_table.items():
        if keyword != "normalization":
            if (tpl is not None) and (tpl != (None, None)):
                total_fractions += tpl[0]
            else:
                remainder_element = keyword
    # print(f"Total fractions {total_fractions}, remainder element {remainder_element}")
    if remainder_element is None:
        raise ValueError("Composition list inconsistent because either fractions for \
                         elements do not add up to 100. or no symbol for matrix defined!")

    if composition_table:  # means != {}
        composition_table[remainder_element] = (1.0e2 - total_fractions, None)
        # error propagation model required

    # document if reporting as percent or fractional values
    composition_table["percent"] = True

    return composition_table



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""A writer for writing dispersion models into a corresponding NXdata entry"""
import re
from typing import Set, Any, Dict
import numpy as np
from scipy.interpolate import interp1d

from pynxtools.dataconverter.readers.rii_database.formula_parser.parser import (
    transformation_formula_parser,
)


class NXdataWriter:
    """Extracts/calculates data and writes it to the NXdata entry"""

    def __init__(
        self,
        entries: Dict[str, Any],
        dispersion_path: str,
        single_params: Dict[str, Dict[str, float]],
        repeated_params: Dict[str, Dict[str, np.ndarray]],
    ):
        self.entries = entries
        self.dispersion_path = dispersion_path
        self.single_params = single_params
        self.repeated_params = repeated_params

        self.table_dispersions = self.get_dispersion_names("DISPERSION_TABLE")
        self.func_dispersions = self.get_dispersion_names("DISPERSION_FUNCTION")

    def get_dispersion_names(self, group_name: str):
        """Get all dispersion names of a group"""
        names = set()
        for entry in self.entries:
            if entry.startswith(f"{self.dispersion_path}/{group_name}"):
                names.add(re.search(rf"{group_name}\[([^\]]+)\]", entry).group(1))

        return names

    def get_wavelength_index(self, table_dispersion: str):
        """Get the wavelength and refractive index for a tabular dispersion"""
        wavelength = self.entries[
            f"{self.dispersion_path}/DISPERSION_TABLE[{table_dispersion}]/wavelength"
        ]
        refractive_index = self.entries[
            f"{self.dispersion_path}/DISPERSION_TABLE[{table_dispersion}]/refractive_index"
        ]

        return wavelength, refractive_index

    def get_index(self, wavelength: np.ndarray, func_dispersion: str):
        """
        Get the refractive index of a functional dispersion for a given wavelength array
        """
        x_axis_name = self.entries[
            (
                f"{self.dispersion_path}/DISPERSION_FUNCTION[{func_dispersion}]"
                "/wavelength_identifier"
            )
        ]
        dfpath = f"{self.dispersion_path}/DISPERSION_FUNCTION[{func_dispersion}]"
        formula = self.entries[f"{dfpath}/formula"]

        return transformation_formula_parser(
            x_axis_name,
            wavelength,
            self.single_params.get(dfpath, {}),
            self.repeated_params.get(dfpath, {}),
        ).parse(formula)

    def add_func_dispersions(self, wavelength: np.ndarray, func_dispersions: Set[str]):
        """Sum the refractive indices of functional dispersions"""
        disp_type, refractive_index = self.get_index(wavelength, func_dispersions.pop())
        for func_dispersion in func_dispersions:
            disp_type2, refractive_index2 = self.get_index(wavelength, func_dispersion)
            if disp_type2 != disp_type:
                raise ValueError("Incompatible dispersions")
            refractive_index += refractive_index2

        return disp_type, refractive_index

    def get_wavelength_range(
        self, func_dispersions: Set[str], no_points: int = 500
    ) -> np.ndarray:
        """Get the provided or default wavelength range for a functional dispersion"""
        min_wavelength = -np.inf
        max_wavelength = np.inf
        for func_dispersion in func_dispersions:
            base_path = f"{self.dispersion_path}/DISPERSION_FUNCTION[{func_dispersion}]"
            if f"{base_path}/wavelength_min" in self.entries:
                new_min = self.entries[f"{base_path}/wavelength_min"]
                min_wavelength = new_min if new_min > min_wavelength else min_wavelength

            if f"{base_path}/wavelength_max" in self.entries:
                new_max = self.entries[f"{base_path}/wavelength_max"]
                max_wavelength = new_max if new_max < max_wavelength else max_wavelength

        return np.linspace(
            0.4 if min_wavelength < 0 else min_wavelength,
            1.0 if max_wavelength == np.inf else max_wavelength,
            no_points,
        )

    def write_nx_data(self, wavelength: np.ndarray, refractive_index: np.ndarray):
        """Writes an NXdata group"""
        path = self.dispersion_path
        self.entries[f"{path}/@default"] = "plot"
        self.entries[f"{path}/plot/@axes"] = "wavelength"
        self.entries[f"{path}/plot/@signal"] = "refractive_index"
        self.entries[f"{path}/plot/@auxiliary_signals"] = "extinction"
        self.entries[f"{path}/plot/wavelength"] = wavelength
        self.entries[f"{path}/plot/wavelength/@units"] = "micrometer"
        self.entries[f"{path}/plot/refractive_index"] = refractive_index.real
        self.entries[f"{path}/plot/extinction"] = refractive_index.imag

    def write_func_dispersions_entries(self):
        """Write the functional dispersion entries to the dict"""
        wavelength = self.get_wavelength_range(self.func_dispersions)
        disp_type, refractive_index = self.add_func_dispersions(
            wavelength, self.func_dispersions
        )

        if disp_type == "eps":
            refractive_index = np.emath.sqrt(refractive_index)

        self.write_nx_data(wavelength, refractive_index)

    def write_table_func_dispersions_entries(self):
        """Write the tabular and functional dispersion entries to the dict"""
        wavelength, refractive_index = self.get_wavelength_index(
            self.table_dispersions.pop()
        )
        disp_type, refractive_index_func = self.add_func_dispersions(
            wavelength, self.func_dispersions
        )

        if disp_type == "eps":
            raise ValueError(
                "Incompatible dispersions detected: Tabular and eps-based function"
            )

        self.write_nx_data(wavelength, refractive_index + refractive_index_func)

    def write_table_dispersions_entries(self):
        """Write tabular dispersion entries to the dict"""
        wlen, rindex = self.get_wavelength_index(self.table_dispersions.pop())
        min_wlen = min(wlen)
        max_wlen = max(wlen)
        interpolations = [interp1d(wlen, rindex)]
        for table_dispersion in self.table_dispersions:
            wlen, rindex = self.get_wavelength_index(table_dispersion)
            min_wlen = min(wlen) if min(wlen) > min_wlen else min_wlen
            max_wlen = max(wlen) if max(wlen) < max_wlen else max_wlen
            interpolations.append(interp1d(wlen, rindex))

        wavelength = np.linspace(min_wlen, max_wlen, 500)
        refractive_index = np.zeros(wavelength.shape, np.cdouble)
        for interp in interpolations:
            refractive_index += interp(wavelength)

        self.write_nx_data(wavelength, refractive_index)

    def write_entries(self):
        """Write the NXdata entries to the dict"""
        if len(self.table_dispersions) == 1 and self.func_dispersions:
            self.write_table_func_dispersions_entries()
            return self.entries

        if len(self.table_dispersions) == 0 and self.func_dispersions:
            self.write_func_dispersions_entries()
            return self.entries

        if self.func_dispersions:
            raise ValueError("Unexpected number of dispersions.")

        if not self.table_dispersions:
            raise ValueError("No valid dispersions found.")

        self.write_table_dispersions_entries()
        return self.entries



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Parse three-dimensional EBSD data from a DREAM3D file."""

# pylint: disable=no-member,too-many-instance-attributes

# type: ignore

# from typing import Dict, Any, List, Tuple

import yaml

import numpy as np

import h5py

# import imageio.v3 as iio
from PIL import Image as pil

# from pynxtools.dataconverter.readers.em_om.utils.dream3d_filter_class \
#     import DreamThreedFilter

from pynxtools.dataconverter.readers.em_om.utils.image_transform import thumbnail

from pynxtools.dataconverter.readers.em_om.utils.em_nexus_plots import HFIVE_WEB_MAX_SIZE


class NxEmOmDreamThreedEbsdParser:
    """Parse DREAM3D EBSD data.

    """

    def __init__(self, file_name, entry_id):
        """Class wrapping dream3d parser."""
        # this is one example which should be extended to interpretation of more
        # DREAM3D filter, currently though all these filter and the DREAM.3D software
        # experiences major refactoring associated with the introduction of DREAM3DNX
        self.file_name = file_name
        self.entry_id = entry_id
        self.pipeline = {}
        # the dictionary nest in DREAM.3D SIMPL pipeline notation
        self.stack_meta = {}
        self.stack = []
        self.roi = []
        self.xyz = {}
        self.phase_id = []
        self.phases = {}

    def parse_pipeline(self):
        """Parse and understand pipeline json description of the dream3d file."""
        h5r = h5py.File(self.file_name, "r")
        dsnm = "Pipeline/Pipeline"
        if dsnm in h5r:
            self.pipeline = yaml.safe_load(h5r[dsnm][()].decode("utf-8"))
        h5r.close()

        print(self.pipeline)
        # add all logical understanding of the pipeline steps
        # ##MK::should be extended

    def parse_roi_geometry(self):
        """Parse the geometry of the 3D dataset ##MK::assuming that it exists."""
        # NEW ISSUE: this is an example how to parse 3D reconstructed EBSD data
        # from the older version of DREAM.3D (not for DREAM.3D NX) maps on NXem_ebsd

        h5r = h5py.File(self.file_name, "r")

        grpnm = "/DataContainers/Small IN100/_SIMPL_GEOMETRY"
        # NEW ISSUE: interpret from filter!
        has_required_metadata = True
        has_correct_shape = True
        req_field_names = [("dims", "DIMENSIONS"),
                           ("origin", "ORIGIN"),
                           ("spacing", "SPACING")]
        for field_tuple in req_field_names:
            if f"{grpnm}/{field_tuple[1]}" in h5r:
                self.stack_meta[field_tuple[0]] = h5r[f"{grpnm}/{field_tuple[1]}"][...]
                if np.shape(self.stack_meta[field_tuple[0]]) != (3,):
                    has_correct_shape = False
            else:
                has_required_metadata = False

        if (has_required_metadata is True) and (has_correct_shape is True):
            grpnm = "/DataContainers/Small IN100/EBSD Scan Data"

            dsnm = f"{grpnm}/Phases"
            self.phase_id = np.asarray(h5r.get(dsnm)[...], np.uint32)
            # DREAM.3D stores int32_t but only >= 0 ##?, we can silently promote to uint32

            dsnm = f"{grpnm}/Image Quality"
            self.roi = np.asarray(h5r.get(dsnm)[...], np.float32)

            dsnm = f"{grpnm}/IPFColor"
            self.stack = np.asarray(h5r.get(dsnm)[...], np.uint8)  # RGB triplet
            # if len(np.shape(self.stack)) == 4:
            #     for i in np.arange(0, 3):
            #         if np.shape(self.stack)[i] >= 1:
            # if np.shape(self.stack)[3] == 3:
            print(np.shape(self.roi))
            print(np.shape(self.phase_id))
            print(np.shape(self.stack))

            # NEW ISSUE: more consistence checks that sizes and shapes match

        h5r.close()

    def parse_roi_dimension_scale_axes(self):
        """Create correctly calibrated dimension scale axes center-of-mass arrays."""
        axes_names = [(0, "x"), (1, "y"), (2, "z")]
        self.xyz = {}
        for axisname in axes_names:
            i = axisname[0]
            # self.xyz[axisname[1]] = np.asarray(np.linspace(self.stack_meta["origin"][i],
            #     self.stack_meta["origin"][i] + self.stack_meta["dims"][i] \
            #     * self.stack_meta["spacing"][i], num=self.stack_meta["dims"][i],
            #     endpoint=True), np.float32)  # DREAM.3D uses single precision
            self.xyz[axisname[1]] = np.asarray(
                np.linspace(self.stack_meta["origin"][i],
                            self.stack_meta["origin"][i] + self.stack_meta["dims"][i]
                            * self.stack_meta["spacing"][i], num=self.stack_meta["dims"][i],
                            endpoint=True),
                np.float32)  # DREAM.3D uses single precision

        # endpoint true? voxel center or its min or max bound?

    def parse_roi_default_plot(self, template: dict) -> dict:
        """Create default plot for the region-of-interest."""
        trg = f"/ENTRY[entry{self.entry_id}]/correlation/region_of_interest/roi"
        template[f"{trg}/title"] = str("Region-of-interest overview image")
        template[f"{trg}/@signal"] = "data"
        # template[f"{trg}/@axes"] = ["axis_x", "axis_y", "axis_z"]
        template[f"{trg}/@axes"] = ["axis_z", "axis_y", "axis_x"]
        template[f"{trg}/@AXISNAME_indices[axis_x_indices]"] = np.uint32(0)
        template[f"{trg}/@AXISNAME_indices[axis_y_indices]"] = np.uint32(1)
        template[f"{trg}/@AXISNAME_indices[axis_z_indices]"] = np.uint32(2)

        trg = f"/ENTRY[entry{self.entry_id}]/correlation/region_of_interest/roi/data"
        template[f"{trg}"] = {"compress": self.roi[:, :, :, 0], "strength": 1}

        template[f"{trg}/@long_name"] = "Signal"
        template[f"{trg}/@CLASS"] = "IMAGE"  # required by H5Web to plot RGB maps
        template[f"{trg}/@IMAGE_VERSION"] = "1.2"
        template[f"{trg}/@SUBCLASS_VERSION"] = np.int64(15)

        axes_names = [("axis_x", "x"), ("axis_y", "y"), ("axis_z", "z")]
        for axisname in axes_names:
            trg = f"/ENTRY[entry{self.entry_id}]/correlation/region_of_interest" \
                  f"/roi/AXISNAME[{axisname[0]}]"
            template[f"{trg}"] = {"compress": self.xyz[axisname[1]], "strength": 1}
            template[f"{trg}/@long_name"] \
                = f"Calibrated position along {axisname[1]}-axis (µm)"
            template[f"{trg}/@units"] = "µm"  # parse from filter!

        return template

    def parse_phases(self, template: dict) -> dict:
        """Parse data for the crystal structure models aka phases."""
        h5r = h5py.File(self.file_name, "r")
        grpnm = "/DataContainers/Small IN100/Phase Data"  # NEW ISSUE: interpret from filter!
        dsnm = f"{grpnm}/CrystalStructures"
        phase_ids = h5r[dsnm][...].flatten()
        print(phase_ids)
        dsnm = f"{grpnm}/MaterialName"
        phase_names = []
        for phase_name in h5r[dsnm][...]:
            phase_names.append(phase_name.decode("utf-8"))
        print(phase_names)
        dsnm = f"{grpnm}/LatticeConstants"
        unit_cells = h5r[dsnm][...]
        identifier = 0
        if np.shape(phase_ids)[0] == np.shape(phase_names)[0]:
            if np.shape(phase_ids)[0] == np.shape(unit_cells)[0]:
                if np.shape(unit_cells)[1] == 6:
                    # NEW ISSUE: assume first one is invalid
                    identifier = 1

        if identifier == 1:  # NEW ISSUE: extend for arbitrary number of phases
            self.phases[identifier] = {}
            self.phases[identifier]["name"] = phase_names[identifier]

            trg = f"/ENTRY[entry{self.entry_id}]/correlation" \
                  f"/EM_EBSD_CRYSTAL_STRUCTURE_MODEL" \
                  f"[em_ebsd_crystal_structure_model{identifier}]"
            template[f"{trg}/phase_identifier"] = np.uint32(identifier)
            template[f"{trg}/phase_name"] = str(phase_names[identifier])
            template[f"{trg}/unit_cell_abc"] = np.asarray(
                [unit_cells[identifier, 0] * 0.1,
                 unit_cells[identifier, 1] * 0.1,
                 unit_cells[identifier, 2] * 0.1], np.float32)
            # ##? DREAM.3D reports in angstroem but no units attribute in dream3d file!
            template[f"{trg}/unit_cell_abc/@units"] = "nm"
            template[f"{trg}/unit_cell_alphabetagamma"] = np.asarray(
                [unit_cells[identifier, 3],
                 unit_cells[identifier, 4],
                 unit_cells[identifier, 5]], np.float32)
            # ##? DREAM.3D reports in degree
            template[f"{trg}/unit_cell_alphabetagamma/@units"] = "°"

        return template

    def parse_inverse_pole_figures(self, template: dict) -> dict:
        """Parse inverse pole figures (IPF) mappings."""
        print("Parse inverse pole figures (IPFs)...")
        # identifier match because phase is a list asc. sorted by numerical keys
        # "first phase_id" in DREAM.3D and em_om is 1
        for identifier in np.arange(1, np.max(self.phase_id) + 1):
            self.parse_inverse_pole_figure_map(identifier, template)
            self.parse_inverse_pole_figure_color_key(identifier, template)
            identifier += 1

        return template

    def parse_inverse_pole_figure_map(self, identifier, template: dict) -> dict:
        """Create default plot for the IPF-Z orientation mapping."""
        phase_id = identifier
        phase_name = self.phases[identifier]["name"]
        print(f"Generate inverse pole figure (IPF) map for {identifier}, {phase_name}...")

        trg = f"/ENTRY[entry{self.entry_id}]/correlation/PROCESS[ipf_map{phase_id}]"
        template[f"{trg}/bitdepth"] = np.uint32(8)
        template[f"{trg}/phase_identifier"] = np.uint32(phase_id)
        template[f"{trg}/phase_name"] = str(phase_name)
        template[f"{trg}/PROGRAM[program1]/program"] = str("dream3d")
        template[f"{trg}/PROGRAM[program1]/program/@version"] = "v6.5.163"
        template[f"{trg}/projection_direction"] = np.asarray([0., 0., 1.], np.float32)

        trg = f"/ENTRY[entry{self.entry_id}]/correlation" \
              f"/PROCESS[ipf_map{phase_id}]/ipf_rgb_map"
        template[f"{trg}/title"] = str("DREAM.3D ROI inverse-pole-figure colored")
        template[f"{trg}/@signal"] = "data"
        # template[f"{trg}/@axes"] = ["axis_x", "axis_y", "axis_z"]
        template[f"{trg}/@axes"] = ["axis_z", "axis_y", "axis_x"]
        template[f"{trg}/@AXISNAME_indices[axis_x_indices]"] = np.uint32(0)
        template[f"{trg}/@AXISNAME_indices[axis_y_indices]"] = np.uint32(1)
        template[f"{trg}/@AXISNAME_indices[axis_z_indices]"] = np.uint32(2)
        # check again order x, y, z ??

        trg = f"/ENTRY[entry{self.entry_id}]/correlation" \
              f"/PROCESS[ipf_map{phase_id}]/ipf_rgb_map/data"
        # NEW ISSUE: needs one more check if precise currently
        ipf_map = np.asarray(np.zeros(np.shape(self.stack), np.uint8))  # 4d array
        # msk = self.phase_id == phase_id
        indices = np.shape(self.stack)
        for i in np.arange(0, indices[0]):
            for j in np.arange(0, indices[1]):
                for k in np.arange(0, indices[2]):
                    if self.phase_id[i, j, k, 0] == phase_id:
                        ipf_map[i, j, k, 0:3] = self.stack[i, j, k, 0:3]
        # likely this can be done with much fancy numpy nd indexing and more pythonic
        # instead of this loop nest but this is right at least...
        # needs a multi-dimensional masking of all values [x, y, z] check order...
        # where self.phase_id[x, y, z] != phase_id set ipf_map[x, y, z, 0:3]
        template[f"{trg}"] = {"compress": ipf_map, "strength": 1}

        template[f"{trg}/@long_name"] = "Signal"
        template[f"{trg}/@CLASS"] = "IMAGE"  # required by H5Web to plot RGB maps
        template[f"{trg}/@IMAGE_VERSION"] = "1.2"
        template[f"{trg}/@SUBCLASS_VERSION"] = np.int64(15)

        axes_names = [("axis_x", "x"), ("axis_y", "y"), ("axis_z", "z")]
        for axisname in axes_names:
            trg = f"/ENTRY[entry{self.entry_id}]/correlation" \
                  f"/PROCESS[ipf_map{phase_id}]/ipf_rgb_map/AXISNAME[{axisname[0]}]"
            template[f"{trg}"] = {"compress": self.xyz[axisname[1]], "strength": 1}
            template[f"{trg}/@long_name"] \
                = f"Calibrated position along {axisname[1]}-axis (µm)"
            template[f"{trg}/@units"] = "µm"  # parse from filter!

        return template

    def parse_inverse_pole_figure_color_key(self, identifier, template: dict) -> dict:
        """Parse color key renderings of inverse-pole-figure (IPF) mappings."""
        # in DREAM.3D not_indexed phase is 0 and "first" phase has ID 1
        # this is coincidently the convention which also this nomad-parser-nexus
        # em_om reader uses
        phase_id = identifier
        phase_name = self.phases[identifier]["name"]
        print(f"Parse inverse pole figure (IPF) color key {identifier}, {phase_name}...")

        # the key problem is that the DREAM.3D pipeline does not store the point group
        # so we need to have a heuristic approach which selects the correct IPF
        # color maps from the available default color mappings
        # https://github.com/BlueQuartzSoftware/DREAM3D/tree/v6_5_163/Source/Plugins/
        # OrientationAnalysis/Data/OrientationAnalysis/IPF_Legend
        # DREAM.3D stores a library of prerendered color keys as image files
        color_key_path = (__file__).replace(
            "dream3d_ebsd_parser.py", "dream3d_v65163_color_keys")
        color_key_file_name = f"{color_key_path}/Cubic_High.png"
        # NEW ISSUE:must not be Cubic_High.png only, this holds only for this example!

        # constraint further to 8bit RGB and no flipping
        # im = np.asarray(imageio.v3.imread(symm_name))
        img = np.asarray(thumbnail(pil.open(color_key_file_name, "r", ["png"]),
                         size=HFIVE_WEB_MAX_SIZE), np.uint8)
        img = img[:, :, 0:3]  # discard potential alpha channel
        # ##MK::need to constrain more the writing of the image that it is guaranteed
        # a specific type of image and bitdepth and color model, and avoid implicit
        # image transformations such as flips or rotations

        trg = f"/ENTRY[entry{self.entry_id}]/correlation" \
              f"/PROCESS[ipf_map{phase_id}]/ipf_rgb_color_model"
        template[f"{trg}/title"] = str("Inverse pole figure color key with SST")
        template[f"{trg}/@signal"] = "data"
        template[f"{trg}/@axes"] = ["axis_y", "axis_x"]
        template[f"{trg}/@AXISNAME_indices[axis_x_indices]"] = np.uint32(0)
        template[f"{trg}/@AXISNAME_indices[axis_y_indices]"] = np.uint32(1)
        trg = f"/ENTRY[entry{self.entry_id}]/correlation" \
              f"/PROCESS[ipf_map{phase_id}]/ipf_rgb_color_model/DATA[data]"
        template[f"{trg}"] = {"compress": img, "strength": 1}
        template[f"{trg}/@CLASS"] = "IMAGE"
        template[f"{trg}/@IMAGE_VERSION"] = "1.2"
        template[f"{trg}/@SUBCLASS_VERSION"] = np.int64(15)

        axes_names = [("axis_y", 0, "y-axis"), ("axis_x", 1, "x-axis")]
        for axis in axes_names:
            trg = f"/ENTRY[entry{self.entry_id}]/correlation/PROCESS[ipf_map{phase_id}]" \
                  f"/ipf_rgb_color_model/AXISNAME[{axis[0]}]"
            template[f"{trg}"] = {"compress": np.asarray(
                np.linspace(1, np.shape(img)[axis[1]], num=np.shape(img)[axis[1]],
                            endpoint=True), np.uint32), "strength": 1}
            template[f"{trg}/@long_name"] = f"Pixel along {axis[2]}"
            template[f"{trg}/@units"] = "px"

        return template

    def parse(self, template: dict) -> dict:
        """Parse NOMAD OASIS relevant data and metadata from a DREAM.3D file."""
        print("Parsing EBSD data from DREAM.3D...")
        print(self.file_name)
        print(f"{self.entry_id}")
        self.parse_pipeline()
        self.parse_roi_geometry()
        self.parse_roi_dimension_scale_axes()
        self.parse_roi_default_plot(template)
        self.parse_phases(template)
        self.parse_inverse_pole_figures(template)
        return template



import os

from click.testing import CliRunner

from ..nyaml2nxdl import nyaml2nxdl as conv
from ..utils.nxdl_utils import find_definition_file

# import subprocess


def test_conversion():
    root = find_definition_file("NXentry")
    # subprocess.run(["python3","-m","dev_tools.nyaml2nxdl.nyaml2nxdl","--input-file",root])
    result = CliRunner().invoke(conv.launch_tool, ["--input-file", root])
    assert result.exit_code == 0
    yaml = root[:-9] + "_parsed.yaml"
    # subprocess.run(["python3","-m","dev_tools.nyaml2nxdl.nyaml2nxdl","--input-file",yaml])
    result = CliRunner().invoke(conv.launch_tool, ["--input-file", yaml])
    assert result.exit_code == 0
    new_root = yaml[:-4] + "nxdl.xml"
    with open(root, encoding="utf-8", mode="r") as tmp_f:
        root_content = tmp_f.readlines()
    with open(new_root, encoding="utf-8", mode="r") as tmp_f:
        new_root_content = tmp_f.readlines()
    assert root_content == new_root_content
    os.remove(yaml)
    os.remove(new_root)



class XMLSyntaxError(Exception):
    """XML file has a syntax error"""

    pass


class NXDLSyntaxError(XMLSyntaxError):
    """XML file written in NXDL has a syntax error"""

    pass


class NXDLParseError(Exception):
    """XML file written in NXDL cannot be parsed"""



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Generic parser for loading atom probe microscopy data into NXapm."""

# pylint: disable=no-member

from typing import Tuple, Any

from pynxtools.dataconverter.readers.base.reader import BaseReader

from pynxtools.dataconverter.readers.apm.utils.apm_define_io_cases \
    import ApmUseCaseSelector

from pynxtools.dataconverter.readers.apm.utils.apm_load_deployment_specifics \
    import NxApmNomadOasisConfigurationParser

from pynxtools.dataconverter.readers.apm.utils.apm_load_generic_eln \
    import NxApmNomadOasisElnSchemaParser

from pynxtools.dataconverter.readers.apm.utils.apm_load_reconstruction \
    import ApmReconstructionParser

from pynxtools.dataconverter.readers.apm.utils.apm_load_ranging \
    import ApmRangingDefinitionsParser

from pynxtools.dataconverter.readers.apm.utils.apm_create_nx_default_plots \
    import apm_default_plot_generator

from pynxtools.dataconverter.readers.apm.utils.apm_generate_synthetic_data \
    import ApmCreateExampleData

# this apm parser combines multiple sub-parsers
# so we need the following input:
# logical analysis which use case
# data input from an ELN (e.g. NOMAD OASIS)
# data input from technology partner files
# functionalities for creating default plots
# developer functionalities for creating synthetic data

# each reconstruction should be stored as an own file because for commercial
# atom probe microscopes it is currently impossible to get less processed data
# from the microscopes
# for development purposes synthetic datasets can be created which are
# for now stored all in the same file. As these use the same dictionary
# the template variable analyses of files larger than the physical main memory
# can currently not be handled


class ApmReader(BaseReader):
    """Parse content from community file formats.

    Specifically, (local electrode) atom probe microscopy and field-ion microscopy
    towards a NXapm.nxdl-compliant NeXus file.

    """

    # pylint: disable=too-few-public-methods

    # Whitelist for the NXDLs that the reader supports and can process
    supported_nxdls = ["NXapm"]

    def read(self,
             template: dict = None,
             file_paths: Tuple[str] = None,
             objects: Tuple[Any] = None) -> dict:
        """Read data from given file, return filled template dictionary apm."""
        template.clear()

        n_entries = 1
        entry_id = 1
        if len(file_paths) == 1:
            if file_paths[0].startswith("synthesize"):
                synthesis_id = int(file_paths[0].replace("synthesize", ""))
                print(f"synthesis_id {synthesis_id}")
            else:
                synthesis_id = 1
            print("Create one synthetic entry in one NeXus file...")
            synthetic = ApmCreateExampleData(synthesis_id)
            synthetic.synthesize(template)
        else:  # eln_data, and ideally recon and ranging definitions from technology partner file
            print("Parse ELN and technology partner file(s)...")
            case = ApmUseCaseSelector(file_paths)
            assert case.is_valid is True, \
                "Such a combination of input-file(s, if any) is not supported !"

            print("Parse (meta)data coming from an ELN...")
            if len(case.eln) == 1:
                nx_apm_eln = NxApmNomadOasisElnSchemaParser(case.eln[0], entry_id)
                nx_apm_eln.report(template)
            else:
                print("No input file defined for eln data !")
                return {}

            print("Parse (meta)data coming from a configuration that specific OASIS...")
            if len(case.cfg) == 1:
                nx_apm_cfg = NxApmNomadOasisConfigurationParser(case.cfg[0], entry_id)
                nx_apm_cfg.report(template)
            # having and or using a deployment-specific configuration is optional

            print("Parse (numerical) data and metadata from ranging definitions file...")
            if len(case.reconstruction) == 1:
                nx_apm_recon = ApmReconstructionParser(case.reconstruction[0], entry_id)
                nx_apm_recon.report(template)
            else:
                print("No input-file defined for reconstructed dataset!")
                return {}
            if len(case.ranging) == 1:
                nx_apm_range = ApmRangingDefinitionsParser(case.ranging[0], entry_id)
                nx_apm_range.report(template)
            else:
                print("No input-file defined for ranging definitions!")
                return {}

        print("Create NeXus default plottable data...")
        apm_default_plot_generator(template, n_entries)

        # print("Reporting state of template before passing to HDF5 writing...")
        # for keyword in template.keys():
        #     print(keyword)
        #     print(template[keyword])

        print("Forward instantiated template to the NXS writer...")
        return template


# This has to be set to allow the convert script to use this reader.
READER = ApmReader



from pathlib import Path
from typing import Iterator
from typing import Optional
from typing import Tuple

from ..globals.directories import get_build_root
from ..globals.directories import get_nxdl_root


def iter_definitions(*subdirs: Tuple[str]) -> Iterator[Path]:
    """Yield all NeXus class definitions"""
    root = get_nxdl_root()
    build_root = get_build_root()
    for subdir in subdirs:
        root /= subdir
    for path in sorted(root.rglob("*.nxdl.xml")):
        try:
            path.relative_to(build_root)
            # Skip file in the build root
            continue
        except ValueError:
            yield path


def find_definition(nxclass: str, *subdirs: Tuple[str]) -> Optional[Path]:
    """Find the definition of a NeXus class"""
    s = nxclass.lower() + ".nxdl"
    for path in iter_definitions(*subdirs):
        if path.stem.lower() == s:
            return path



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Functions for reading metadata values from Perkin Ellmer files"""
from datetime import datetime

# The min & max wavelength the instrument can measure
MIN_WAVELENGTH = 190.
MAX_WAVELENGTH = 3350.


def read_start_date(metadata: list) -> str:
    """Reads the start date from the metadata"""
    century = str(datetime.now().year // 100)
    formated_date = metadata[3].replace("/", "-")
    return f"{century}{formated_date}T{metadata[4]}000Z"


def read_sample_attenuator(metadata: list) -> int:
    """Reads the sample attenuator from the metadata"""
    return int(metadata[47].split()[0].split(":")[1])


def read_ref_attenuator(metadata: list) -> int:
    """Reads the sample attenuator from the metadata"""
    return int(metadata[47].split()[1].split(":")[1])


def is_depolarizer_on(metadata: list) -> bool:
    """Reads whether the depolarizer was active during the measurement"""
    return metadata[46] == 'on'


def read_uv_monochromator_range(metadata: list) -> list:
    """Reads the uv monochromator range from the metadata"""
    monochromator_change = float(metadata[41])
    return [MIN_WAVELENGTH, monochromator_change]


def read_visir_monochromator_range(metadata: list) -> list:
    """Reads the visir monochromator range from the metadata"""
    monochromator_change = float(metadata[41])
    return [monochromator_change, MAX_WAVELENGTH]


def get_d2_range(metadata: list) -> list:
    """Reads the D2 lamp range from the metadata"""
    lamp_change = float(metadata[42])
    return [MIN_WAVELENGTH, lamp_change]


def get_halogen_range(metadata: list) -> list:
    """Reads the halogen lamp range from the metadata"""
    lamp_change = float(metadata[42])
    return [lamp_change, MAX_WAVELENGTH]



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Test cases for template class for the DataConverter"""

from .test_helpers import fixture_template  # pylint: disable=unused-import


def test_rename_entry(template):
    """Unit test for the rename entry function"""
    template.rename_entry("entry", "newentry")
    assert "/ENTRY[entry]/program_name" not in template.keys()


def test_add_entry(template):
    """Unit test for adding an entry to the template class."""
    template.add_entry("test_entry")
    assert "/ENTRY[entry]/program_name" in template.keys()
    assert "/ENTRY[test_entry]/program_name" in template.keys()



from pathlib import Path
from nexusformat.nexus import nxopen

filename = str(
    Path(__file__).absolute().parent.parent
    / "simple_example_basic"
    / "simple_example_basic.nexus.hdf5"
)
with nxopen(filename) as f:
    # find the default NXdata group
    nx_data = f.get_default()
    signal = nx_data.nxsignal
    axes = nx_data.nxaxes[0]

nx_data.plot() # plot the data using Matplotlib

print(f"file: {f.nxfilename}")
print(f"signal: {signal.nxname}")
print(f"axes: {axes.nxname}")
print(f"{axes.nxname} {signal.nxname}")
for x, y in zip(axes, signal):
    print(x, y)



import os
import re
from collections import OrderedDict
from html import parser as HTMLParser
from pathlib import Path
from typing import List
from typing import Optional

import lxml

from ..globals.directories import get_nxdl_root
from ..globals.errors import NXDLParseError
from ..globals.nxdl import NXDL_NAMESPACE
from ..globals.urls import REPO_URL
from ..utils import nxdl_utils as pynxtools_nxlib
from ..utils.github import get_file_contributors_via_api
from ..utils.types import PathLike
from .anchor_list import AnchorRegistry

# controlling the length of progressively more indented sub-node
MIN_COLLAPSE_HINT_LINE_LENGTH = 20
MAX_COLLAPSE_HINT_LINE_LENGTH = 80


class NXClassDocGenerator:
    """Generate documentation in reStructuredText markup
    for a NeXus class definition."""

    _INDENTATION_UNIT = " " * 2

    _CATEGORY_TO_LISTING = {
        "base": "base class",
        "application": "application definition",
    }

    def __init__(self) -> None:
        self._rst_lines = None
        self._reset()

    def _reset(self):
        self._anchor_registry = None
        self._listing_category = None
        self._use_application_defaults = None

    def __call__(
        self, nxdl_file: PathLike, anchor_registry: Optional[AnchorRegistry] = None
    ) -> List[str]:
        self._rst_lines = list()
        self._anchor_registry = anchor_registry
        nxdl_file = Path(nxdl_file)
        if anchor_registry:
            self._anchor_registry.nxdl_file = nxdl_file
        try:
            try:
                self._parse_nxdl_file(nxdl_file)
            except Exception:
                raise NXDLParseError(nxdl_file)
        finally:
            self._reset()
        return self._rst_lines

    def _parse_nxdl_file(self, nxdl_file: Path):
        assert nxdl_file.is_file()
        tree = lxml.etree.parse(str(nxdl_file))
        root = tree.getroot()

        # NXDL_NAMESPACE needs to be a globally unique identifier of
        # the NXDL schema. It needs to match the xmlns attribute
        # in the NXDL definition of the NeXus class.
        ns = {"nx": NXDL_NAMESPACE}

        nxclass_name = root.get("name")
        category = root.attrib["category"]
        title = nxclass_name
        parent_path = "/" + nxclass_name  # absolute path of parent nodes, no trailing /
        if len(nxclass_name) < 2 or nxclass_name[0:2] != "NX":
            raise Exception(
                f'Unexpected class name "{nxclass_name}"; does not start with NX'
            )
        lexical_name = nxclass_name[2:]  # without padding 'NX', for indexing
        self._listing_category = self._CATEGORY_TO_LISTING[category]
        self._use_application_defaults = category == "application"

        # print ReST comments and section header
        source = os.path.relpath(nxdl_file, get_nxdl_root())
        self._print(
            f".. auto-generated by {__name__} from the NXDL source {source} -- DO NOT EDIT"
        )

        self._print("")
        self._print(".. index::")
        self._print(f"    ! {nxclass_name} ({self._listing_category})")
        self._print(f"    ! {lexical_name} ({self._listing_category})")
        self._print(
            f"    see: {lexical_name} ({self._listing_category}); {nxclass_name}"
        )
        self._print("")
        self._print(f".. _{nxclass_name}:\n")
        self._print("=" * len(title))
        self._print(title)
        self._print("=" * len(title))

        # print category & parent class
        extends = root.get("extends")
        if extends is None:
            extends = "none"
        else:
            extends = f":ref:`{extends}`"

        # add the contributors as variables to the rst file that will
        nxdl_root = get_nxdl_root()
        rel_path = str(nxdl_file.relative_to(nxdl_root))
        rel_html = str(rel_path).replace(os.sep, "/")
        contribs_dct = get_file_contributors_via_api("definitions", rel_html)
        if contribs_dct is not None:
            self._print("")
            self._print("..")
            self._print("    Contributors List")
            for date_str, contrib_dct in contribs_dct.items():
                date_str = date_str.split("T")[0]
                name = contrib_dct["name"]
                gh_login_nm = contrib_dct["commit_dct"]["committer"]["login"]
                gh_avatar_url = contrib_dct["commit_dct"]["committer"]["avatar_url"]
                self._print("")
                s = "|".join([name, gh_login_nm, gh_avatar_url, date_str])
                self._print(f"    .. |contrib_name| replace:: {s}")

        self._print("")
        self._print("**Status**:\n")
        self._print(f"  {self._listing_category.strip()}, extends {extends}")

        self._print_if_deprecated(ns, root, "")

        # print official description of this class
        self._print("")
        self._print("**Description**:\n")
        self._print_doc_enum("", ns, root, required=True)

        # print symbol list
        node_list = root.xpath("nx:symbols", namespaces=ns)
        self._print("**Symbols**:\n")
        if len(node_list) == 0:
            self._print("  No symbol table\n")
        elif len(node_list) > 1:
            raise Exception(f"Invalid symbol table in {nxclass_name}")
        else:
            self._print_doc_enum("", ns, node_list[0])
            for node in node_list[0].xpath("nx:symbol", namespaces=ns):
                doc = self._get_doc_line(ns, node)
                self._print(f"  **{node.get('name')}**", end="")
                if doc:
                    self._print(f": {doc}", end="")
                self._print("\n")

        # print group references
        self._print("**Groups cited**:")
        node_list = root.xpath("//nx:group", namespaces=ns)
        groups = []
        for node in node_list:
            g = node.get("type")
            if g.startswith("NX") and g not in groups:
                groups.append(g)
        if len(groups) == 0:
            self._print("  none\n")
        else:
            out = [(f":ref:`{g}`") for g in groups]
            txt = ", ".join(sorted(out))
            self._print(f"  {txt}\n")
            out = [
                ("%s (base class); used in %s" % (g, self._listing_category))
                for g in groups
            ]
            txt = ", ".join(out)
            self._print(f".. index:: {txt}\n")

        # print full tree
        self._print("**Structure**:\n")
        for subnode in root.xpath("nx:attribute", namespaces=ns):
            optional = self._get_required_or_optional_text(subnode)
            self._print_attribute(
                ns, "file", subnode, optional, self._INDENTATION_UNIT, parent_path
            )  # FIXME: +"/"+name )
        self._print_full_tree(
            ns, root, nxclass_name, self._INDENTATION_UNIT, parent_path
        )

        self._print_anchor_list()

        # print NXDL source location
        self._print("")
        self._print("**NXDL Source**:")
        nxdl_root = get_nxdl_root()
        rel_path = str(nxdl_file.relative_to(nxdl_root))
        rel_html = str(rel_path).replace(os.sep, "/")
        self._print(f"  {REPO_URL}/{rel_html}")

        return self._rst_lines

    def _print_anchor_list(self):
        """Print the list of hypertext anchors."""
        if not self._anchor_registry:
            return
        anchors = self._anchor_registry.flush_anchor_buffer()
        if not anchors:
            return

        self._print("")
        self._print("Hypertext Anchors")
        self._print("-----------------\n")
        self._print(
            "List of hypertext anchors for all groups, fields,\n"
            "attributes, and links defined in this class.\n\n"
        )

        def sorter(key):
            return key.lower()

        rst = [f"* :ref:`{ref} <{ref}>`" for ref in sorted(anchors, key=sorter)]

        self._print("\n".join(rst))

    @staticmethod
    def _format_type(node):
        typ = node.get("type", ":ref:`NX_CHAR <NX_CHAR>`")  # per default
        if typ.startswith("NX_"):
            typ = f":ref:`{typ} <{typ}>`"
        return typ

    @staticmethod
    def _format_units(node):
        units = node.get("units", "")
        if not units:
            return ""
        if units.startswith("NX_"):
            units = rf"\ :ref:`{units} <{units}>`"
        return f" {{units={units}}}"

    @staticmethod
    def _get_doc_blocks(ns, node):
        docnodes = node.xpath("nx:doc", namespaces=ns)
        if docnodes is None or len(docnodes) == 0:
            return ""
        if len(docnodes) > 1:
            raise Exception(
                f"Too many doc elements: line {node.sourceline}, {Path(node.base).name}"
            )
        docnode = docnodes[0]

        # be sure to grab _all_ content in the documentation
        # it might look like XML
        s = lxml.etree.tostring(
            docnode, pretty_print=True, method="c14n", with_comments=False
        ).decode("utf-8")
        m = re.search(r"^<doc[^>]*>\n?(.*)\n?</doc>$", s, re.DOTALL)
        if not m:
            raise Exception(f"unexpected docstring [{s}] ")
        text = m.group(1)

        # substitute HTML entities in markup: "<" for "&lt;"
        # thanks: http://stackoverflow.com/questions/2087370/decode-html-entities-in-python-string
        htmlparser = HTMLParser.HTMLParser()
        try:  # see #661
            import html

            text = html.unescape(text)
        except (ImportError, AttributeError):
            text = htmlparser.unescape(text)

        # Blocks are separated by whitelines
        blocks = re.split("\n\\s*\n", text)
        if len(blocks) == 1 and len(blocks[0].splitlines()) == 1:
            return [blocks[0].rstrip().lstrip()]

        # Indentation must be given by first line
        m = re.match(r"(\s*)(\S+)", blocks[0])
        if not m:
            return [""]
        indent = m.group(1)

        # Remove common indentation as determined from first line
        if indent == "":
            raise Exception(
                "Missing initial indentation in <doc> of %s [%s]"
                % (node.get("name"), blocks[0])
            )

        out_blocks = []
        for block in blocks:
            lines = block.rstrip().splitlines()
            out_lines = []
            for line in lines:
                if line[: len(indent)] != indent:
                    raise Exception(
                        'Bad indentation in <doc> of %s [%s]: expected "%s" found "%s".'
                        % (
                            node.get("name"),
                            block,
                            re.sub(r"\t", "\\\\t", indent),
                            re.sub(r"\t", "\\\\t", line),
                        )
                    )
                out_lines.append(line[len(indent) :])
            out_blocks.append("\n".join(out_lines))
        return out_blocks

    def _get_doc_line(self, ns, node):
        blocks = self._get_doc_blocks(ns, node)
        if len(blocks) == 0:
            return ""
        if len(blocks) > 1:
            raise Exception(f"Unexpected multi-paragraph doc [{'|'.join(blocks)}]")
        return re.sub(r"\n", " ", blocks[0])

    def _get_minOccurs(self, node):
        """
        get the value for the ``minOccurs`` attribute

        :param obj node: instance of lxml.etree._Element
        :returns str: value of the attribute (or its default)
        """
        # TODO: can we improve on the default by examining nxdl.xsd?
        minOccurs_default = str(int(self._use_application_defaults))
        minOccurs = node.get("minOccurs", minOccurs_default)
        return minOccurs

    def _get_required_or_optional_text(self, node):
        """
        make clear if a reported item is required or optional

        :param obj node: instance of lxml.etree._Element
        :returns: formatted text
        """
        tag = node.tag.split("}")[-1]
        if tag in ("field", "group"):
            optional_default = not self._use_application_defaults
            optional = node.get("optional", optional_default) in (True, "true", "1", 1)
            recommended = node.get("recommended", None) in (True, "true", "1", 1)
            minOccurs = self._get_minOccurs(node)
            if recommended:
                optional_text = "(recommended) "
            elif minOccurs in ("0", 0) or optional:
                optional_text = "(optional) "
            elif minOccurs in ("1", 1):
                optional_text = "(required) "
            else:
                # this is unexpected and remarkable
                # TODO: add a remark to the log
                optional_text = f"(``minOccurs={str(minOccurs)}``) "
        elif tag in ("attribute",):
            optional_default = not self._use_application_defaults
            optional = node.get("optional", optional_default) in (True, "true", "1", 1)
            recommended = node.get("recommended", None) in (True, "true", "1", 1)
            optional_text = {True: "(optional) ", False: "(required) "}[optional]
            if recommended:
                optional_text = "(recommended) "
        else:
            optional_text = "(unknown tag: " + str(tag) + ") "
        return optional_text

    def _analyze_dimensions(self, ns, parent) -> str:
        """These are the different dimensions that can occur:

        1. Fixed rank

            <dimensions rank="dataRank">
            <dim index="1" value="a" />
            <dim index="2" value="b" />
            <dim index="3" value="c" />
            </dimensions>

        2. Variable rank because of optional dimensions

            <dimensions rank="dataRank">
            <dim index="1" value="a" />
            <dim index="2" value="b" />
            <dim index="3" value="c" />
            <dim index="4" value="d" required="false"/>
            </dimensions>

        3. Variable rank because no dimensions specified

            <dimensions rank="dataRank">
            </dimensions>

        4. Rank and dimensions equal to that of another field called `field_name`

            <dimensions rank="dataRank">
            <dim index="1" ref="field_name" />
            </dimensions>
        """
        node_list = parent.xpath("nx:dimensions", namespaces=ns)
        if len(node_list) != 1:
            return ""
        node = node_list[0]
        node_list = node.xpath("nx:dim", namespaces=ns)

        dims = []
        optional = False
        for subnode in node_list:
            # Dimension index (starts from index 1)
            index = subnode.get("index", "")
            if not index.isdigit():
                raise RuntimeError("A dimension must have an index")
            index = int(index)
            if index <= 0:
                # No longer permitted
                raise RuntimeError(
                    "A dimension's index must be a positive integer (>=1)"
                )

            # Expand dimensions when needed
            index -= 1
            nadd = max(index - len(dims) + 1, 0)
            if nadd:
                dims += ["."] * nadd

            # Dimension symbol
            dim = subnode.get("value")  # integer or symbol from the table
            if not dim:
                ref = subnode.get("ref")
                if ref:
                    return (
                        f" (Rank: same as field {ref}, Dimensions: same as field {ref})"
                    )
                dim = "."  # dimension has no symbol

            # Dimension might be optional
            if subnode.get("required", "true").lower() == "false":
                optional = True
            elif optional:
                raise RuntimeError(
                    "A required dimension cannot come after an optional dimension"
                )
            if optional:
                dim = f"[{dim}]"

            dims[index] = dim

        # When the rank is missing, set to the number of dimensions when
        # there are dimensions specified and none of them are optional.
        ndims = len(dims)
        rank = node.get("rank", None)
        if rank is None and not optional and ndims:
            rank = str(ndims)

        # Validate rank and dimensions
        rank_is_fixed = rank and rank.isdigit()
        if optional and rank_is_fixed:
            raise RuntimeError("A fixed rank cannot have optional dimensions")
        if rank_is_fixed and ndims and int(rank) != ndims:
            raise RuntimeError(
                "The rank and the number of dimensions do not correspond"
            )

        # Omit rank and/or dimensions when not specified
        if rank and dims:
            dims = ", ".join(dims)
            return f" (Rank: {rank}, Dimensions: [{dims}])"
        elif rank:
            return f" (Rank: {rank})"
        elif dims:
            dims = ", ".join(dims)
            return f" (Dimensions: [{dims}])"
        return ""

    def _hyperlink_target(self, parent_path, name, nxtype):
        """Return internal hyperlink target for HTML anchor."""
        if nxtype == "attribute":
            sep = "@"
        else:
            sep = "/"
        target = f"{parent_path}{sep}{name}-{nxtype}"
        if self._anchor_registry:
            self._anchor_registry.add(target)
        return f".. _{target}:\n"

    def _print_enumeration(self, indent, ns, parent):
        node_list = parent.xpath("nx:item", namespaces=ns)
        if len(node_list) == 0:
            return ""

        if len(node_list) == 1:
            self._print(f"{indent}Obligatory value:", end="")
        else:
            self._print(f"{indent}Any of these values:", end="")

        docs = OrderedDict()
        for item in node_list:
            name = item.get("value")
            docs[name] = self._get_doc_line(ns, item)

        ENUMERATION_INLINE_LENGTH = 60

        def show_as_typed_text(msg):
            return f"``{msg}``"

        oneliner = " | ".join(map(show_as_typed_text, docs.keys()))
        if (
            any(doc for doc in docs.values())
            or len(oneliner) > ENUMERATION_INLINE_LENGTH
        ):
            # print one item per line
            self._print("\n")
            for name, doc in docs.items():
                self._print(f"{indent}  * {show_as_typed_text(name)}", end="")
                if doc:
                    self._print(f": {doc}", end="")
                self._print("\n")
        else:
            # print all items in one line
            self._print(f" {oneliner}")
        self._print("")

    def _print_doc(self, indent, ns, node, required=False):
        blocks = self._get_doc_blocks(ns, node)
        if len(blocks) == 0:
            if required:
                raise Exception("No documentation for: " + node.get("name"))
            self._print("")
        else:
            for block in blocks:
                for line in block.splitlines():
                    self._print(f"{indent}{line}")
                self._print()

    def long_doc(self, ns, node, left_margin):
        length = 0
        line = "documentation"
        fnd = False
        blocks = self._get_doc_blocks(ns, node)
        max_characters = max(
            MIN_COLLAPSE_HINT_LINE_LENGTH, (MAX_COLLAPSE_HINT_LINE_LENGTH - left_margin)
        )
        for block in blocks:
            lines = block.splitlines()
            length += len(lines)
            for single_line in lines:
                if len(single_line) > 2 and single_line[0] != "." and not fnd:
                    fnd = True
                    line = single_line[:max_characters]
        return (length, line, blocks)

    def _print_doc_enum(self, indent, ns, node, required=False):
        collapse_indent = indent
        node_list = node.xpath("nx:enumeration", namespaces=ns)
        (doclen, line, blocks) = self.long_doc(ns, node, len(indent))
        if len(node_list) + doclen > 1:
            collapse_indent = f"{indent}    "
            self._print(f"{indent}{self._INDENTATION_UNIT}.. collapse:: {line} ...\n")
        self._print_doc(
            collapse_indent + self._INDENTATION_UNIT, ns, node, required=required
        )
        if len(node_list) == 1:
            self._print_enumeration(
                collapse_indent + self._INDENTATION_UNIT, ns, node_list[0]
            )

    def _print_attribute(self, ns, kind, node, optional, indent, parent_path):
        name = node.get("name")
        index_name = name
        self._print(
            f"{indent}" f"{self._hyperlink_target(parent_path, name, 'attribute')}"
        )
        self._print(f"{indent}.. index:: {index_name} ({kind} attribute)\n")
        self._print(
            f"{indent}**@{name}**: {optional}{self._format_type(node)}{self._format_units(node)} {self.get_first_parent_ref(f'{parent_path}/{name}', 'attribute')}\n"
        )
        self._print_doc_enum(indent, ns, node)

    def _print_if_deprecated(self, ns, node, indent):
        deprecated = node.get("deprecated", None)
        if deprecated is not None:
            self._print(f"\n{indent}.. index:: deprecated\n")
            self._print(f"\n{indent}**DEPRECATED**: {deprecated}\n")

    def _print_full_tree(self, ns, parent, name, indent, parent_path):
        """
        recursively print the full tree structure

        :param dict ns: dictionary of namespaces for use in XPath expressions
        :param lxml_element_node parent: parent node to be documented
        :param str name: name of elements, such as NXentry/NXuser
        :param indent: to keep track of indentation level
        :param parent_path: NX class path of parent nodes
        """

        self._use_application_defaults = self._listing_category in (
            "application definition",
            "contributed definition",
        )

        for node in parent.xpath("nx:field", namespaces=ns):
            name = node.get("name")
            index_name = name
            dims = self._analyze_dimensions(ns, node)

            optional_text = self._get_required_or_optional_text(node)
            self._print(f"{indent}{self._hyperlink_target(parent_path, name, 'field')}")
            self._print(f"{indent}.. index:: {index_name} (field)\n")
            self._print(
                f"{indent}**{name}**: "
                f"{optional_text}"
                f"{self._format_type(node)}"
                f"{dims}"
                f"{self._format_units(node)}"
                f" {self.get_first_parent_ref(f'{parent_path}/{name}', 'field')}"
                "\n"
            )

            self._print_if_deprecated(ns, node, indent + self._INDENTATION_UNIT)
            self._print_doc_enum(indent, ns, node)

            for subnode in node.xpath("nx:attribute", namespaces=ns):
                optional = self._get_required_or_optional_text(subnode)
                self._print_attribute(
                    ns,
                    "field",
                    subnode,
                    optional,
                    indent + self._INDENTATION_UNIT,
                    parent_path + "/" + name,
                )

        for node in parent.xpath("nx:group", namespaces=ns):
            name = node.get("name", "")
            typ = node.get("type", "untyped (this is an error; please report)")

            optional_text = self._get_required_or_optional_text(node)
            if typ.startswith("NX"):
                if name == "":
                    name = typ.lstrip("NX").upper()
                typ = f":ref:`{typ}`"
            hTarget = self._hyperlink_target(parent_path, name, "group")
            # target = hTarget.replace(".. _", "").replace(":\n", "")
            # TODO: https://github.com/nexusformat/definitions/issues/1057
            self._print(f"{indent}{hTarget}")
            self._print(
                f"{indent}**{name}**: {optional_text}{typ} {self.get_first_parent_ref(f'{parent_path}/{name}', 'group')}\n"
            )

            self._print_if_deprecated(ns, node, indent + self._INDENTATION_UNIT)
            self._print_doc_enum(indent, ns, node)

            for subnode in node.xpath("nx:attribute", namespaces=ns):
                optional = self._get_required_or_optional_text(subnode)
                self._print_attribute(
                    ns,
                    "group",
                    subnode,
                    optional,
                    indent + self._INDENTATION_UNIT,
                    parent_path + "/" + name,
                )

            nodename = "%s/%s" % (name, node.get("type"))
            self._print_full_tree(
                ns,
                node,
                nodename,
                indent + self._INDENTATION_UNIT,
                parent_path + "/" + name,
            )

        for node in parent.xpath("nx:link", namespaces=ns):
            name = node.get("name")
            self._print(f"{indent}{self._hyperlink_target(parent_path, name, 'link')}")
            self._print(
                f"{indent}**{name}**: "
                ":ref:`link<Design-Links>` "
                f"(suggested target: ``{node.get('target')}``)"
                "\n"
            )
            self._print_doc_enum(indent, ns, node)

    def _print(self, *args, end="\n"):
        # TODO: change instances of \t to proper indentation
        self._rst_lines.append(" ".join(args) + end)

    def get_first_parent_ref(self, path, tag):
        nx_name = path[1 : path.find("/", 1)]
        path = path[path.find("/", 1) :]

        try:
            parents = pynxtools_nxlib.get_inherited_nodes(path, nx_name)[2]
        except FileNotFoundError:
            return ""
        if len(parents) > 1:
            parent = parents[1]
            parent_path = parent_display_name = parent.attrib["nxdlpath"]
            parent_path_segments = parent_path[1:].split("/")
            parent_def_name = parent.attrib["nxdlbase"][
                parent.attrib["nxdlbase"]
                .rfind("/") : parent.attrib["nxdlbase"]
                .rfind(".nxdl")
            ]

            # Case where the first parent is a base_class
            if parent_path_segments[0] == "":
                return ""

            # special treatment for NXnote@type
            if (
                tag == "attribute"
                and parent_def_name == "/NXnote"
                and parent_path == "/type"
            ):
                return ""

            if tag == "attribute":
                pos_of_right_slash = parent_path.rfind("/")
                parent_path = (
                    parent_path[:pos_of_right_slash]
                    + "@"
                    + parent_path[pos_of_right_slash + 1 :]
                )
            parent_display_name = f"{parent_def_name[1:]}{parent_path}"
            return f":ref:`⤆ </{parent_display_name}-{tag}>`"
        return ""



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Utility for creating list of numpy arrays with dimension scale axes values."""

# pylint: disable=no-member

from typing import List, Any

import numpy as np

from pynxtools.dataconverter.readers.em_nion.map_concepts.swift_display_items_to_nx \
    import metadata_constraints, check_existence_of_required_fields  # nexus_concept_dict


def get_list_of_dimension_scale_axes(dct: dict) -> list:  # , concept_key: str
    """Create a list of dimension scale axes value, unit tuples."""
    # use only when we know already onto which concept a display_item will be mapped
    axes: List[Any] = []
    if (check_existence_of_required_fields(dct, metadata_constraints) is False):
        return axes
    #        or concept_key not in nexus_concept_dict.keys():
    # if nexus_concept_dict[concept_key] is None:
    #     return axes
    if len(dct["dimensional_calibrations"]) != len(dct["data_shape"]):
        return axes

    print(dct["dimensional_calibrations"])
    print(dct["data_shape"])

    for idx in np.arange(0, len(dct["dimensional_calibrations"])):
        nvalues = dct["data_shape"][idx]
        axis_dict = dct["dimensional_calibrations"][idx]
        if isinstance(nvalues, int) and isinstance(axis_dict, dict):
            if (nvalues > 0) \
                    and (set(axis_dict.keys()) == set(["offset", "scale", "units"])):
                start = axis_dict["offset"] + 0.5 * axis_dict["scale"]
                stop = axis_dict["offset"] + ((nvalues - 1) + 0.5) * axis_dict["scale"]
                axes.append(
                    {"value": np.asarray(np.linspace(start,
                                                     stop,
                                                     num=nvalues,
                                                     endpoint=True), np.float64),
                     "unit": axis_dict["units"]})
    return axes



from contextlib import contextmanager
from typing import Optional

import lxml.etree

from ..globals import errors
from ..globals.directories import get_xsd_file
from ..utils.types import PathLike


def nxdl_schema() -> lxml.etree.XMLSchema:
    return lxml.etree.XMLSchema(lxml.etree.parse(get_xsd_file()))


def validate_definition(
    xml_file_name: PathLike,
    xml_schema: Optional[lxml.etree.XMLSchema] = None,
):
    xml_file_name = str(xml_file_name)
    with _handle_xml_error(xml_file_name, lxml.etree.XMLSyntaxError):
        xml_tree = lxml.etree.parse(xml_file_name)
    if xml_schema is None:
        xml_schema = nxdl_schema()
    with _handle_xml_error(xml_file_name, lxml.etree.DocumentInvalid):
        xml_schema.assertValid(xml_tree)


@contextmanager
def _handle_xml_error(xml_file_name: str, *exception_types):
    try:
        yield
    except exception_types as e:
        raise errors.XMLSyntaxError(
            "\n  " + "\n  ".join([xml_file_name] + str(e).rsplit(":", 1))
        ) from e



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Definition of DREAM3D SIMPL filters and their controlled vocabulary."""

# pylint: disable=no-member,pointless-string-statement


'''
class DreamThreedFilter():
    """Description of a filter/plug-in from DREAM.3D."""

    def __init__(self, version="", name="", **kwargs):
        # mandatory
        self.version = ""
        assert isinstance(version, str), "Argument version has to be a string!"
        # assert version != "", "Version mandatory!"
        self.version = version
        self.name = ""
        assert isinstance(name, str), "Argument name has to be a string!"
        # assert name != "", "Name mandatory!"
        self.name = name

        # optional
        self.human_label = ""
        if "human_label" in kwargs:
            self.human_label = kwargs["human_label"]
        self.enabled = False
        if "enabled" in kwargs:
            self.enabled = kwargs["enabled"]
        self.uuid = ""
        if "uuid" in kwargs:
            self.uuid = kwargs["uuid"]

        # controlled flattened keywords
        self.terms = {}
        # if dct is not None:
        #     for key in dct.keys():
        #         self.terms[key] = "add something useful"
        print(f"DREAM.3D filter instance, version {self.version}, name {self.name}")

    def debug(self):
        """Offer functionality testing"""
        tmp = DreamThreedFilter(version="12", name="test")
        print(f"DEBUG: {tmp.version}")
'''



#!/usr/bin/env python
"""Writes a NeXus HDF5 file using h5py and numpy"""

from pathlib import Path
import datetime
import h5py  # HDF5 support
import numpy

print("Write a NeXus HDF5 file")
fileName = "simple_example_basic.nexus.hdf5"
timestamp = datetime.datetime.now().astimezone().isoformat()

# load data from two column format
data_filename = str(Path(__file__).absolute().parent.parent / "simple_example.dat")
data = numpy.loadtxt(data_filename).T
mr_arr = data[0]
i00_arr = numpy.asarray(data[1], "int32")

# create the HDF5 NeXus file
with h5py.File(fileName, "w") as f:
    # point to the default data to be plotted
    f.attrs["default"] = "entry"
    # give the HDF5 root some more attributes
    f.attrs["file_name"] = fileName
    f.attrs["file_time"] = timestamp
    f.attrs["instrument"] = "APS USAXS at 32ID-B"
    f.attrs["creator"] = "simple_example_basic_write.py"
    f.attrs["NeXus_version"] = "4.3.0"
    f.attrs["HDF5_Version"] = h5py.version.hdf5_version
    f.attrs["h5py_version"] = h5py.version.version

    # create the NXentry group
    nxentry = f.create_group("entry")
    nxentry.attrs["NX_class"] = "NXentry"
    nxentry.attrs["default"] = "mr_scan"
    nxentry.create_dataset("title", data="1-D scan of I00 v. mr")

    # create the NXentry group
    nxdata = nxentry.create_group("mr_scan")
    nxdata.attrs["NX_class"] = "NXdata"
    nxdata.attrs["signal"] = "I00"  # Y axis of default plot
    nxdata.attrs["axes"] = "mr"  # X axis of default plot
    nxdata.attrs["mr_indices"] = [
        0,
    ]  # use "mr" as the first dimension of I00

    # X axis data
    ds = nxdata.create_dataset("mr", data=mr_arr)
    ds.attrs["units"] = "degrees"
    ds.attrs["long_name"] = "USAXS mr (degrees)"  # suggested X axis plot label

    # Y axis data
    ds = nxdata.create_dataset("I00", data=i00_arr)
    ds.attrs["units"] = "counts"
    ds.attrs["long_name"] = "USAXS I00 (counts)"  # suggested Y axis plot label

print("wrote file:", fileName)



#!/usr/bin/env python
"""
Writes a simple NeXus HDF5 file using h5py with links
according to the example from Figure 2.1 in the Design chapter
"""

from pathlib import Path
import h5py
import numpy

filename = str(Path(__file__).absolute().parent.parent / "simple_example.dat")
buffer = numpy.loadtxt(filename).T
tthData = buffer[0]  # float[]
countsData = numpy.asarray(buffer[1], "int32")  # int[]

with h5py.File("simple_example_write2.hdf5", "w") as f:  # create the HDF5 NeXus file
    f.attrs["default"] = "entry"

    nxentry = f.create_group("entry")
    nxentry.attrs["NX_class"] = "NXentry"
    nxentry.attrs["default"] = "data"

    nxinstrument = nxentry.create_group("instrument")
    nxinstrument.attrs["NX_class"] = "NXinstrument"

    nxdetector = nxinstrument.create_group("detector")
    nxdetector.attrs["NX_class"] = "NXdetector"

    # store the data in the NXdetector group
    ds_tth = nxdetector.create_dataset("two_theta", data=tthData)
    ds_tth.attrs["units"] = "degrees"
    ds_counts = nxdetector.create_dataset("counts", data=countsData)
    ds_counts.attrs["units"] = "counts"

    # create the NXdata group to define the default plot
    nxdata = nxentry.create_group("data")
    nxdata.attrs["NX_class"] = "NXdata"
    nxdata.attrs["signal"] = "counts"
    nxdata.attrs["axes"] = "two_theta"
    nxdata.attrs["two_theta_indices"] = [
        0,
    ]

    source_addr = "/entry/instrument/detector/two_theta"  # existing data
    target_addr = "two_theta"  # new location
    ds_tth.attrs["target"] = source_addr  # a NeXus API convention for links
    nxdata[target_addr] = f[source_addr]  # hard link
    # nxdata._id.link(source_addr, target_addr, h5py.h5g.LINK_HARD)

    source_addr = "/entry/instrument/detector/counts"  # existing data
    target_addr = "counts"  # new location
    ds_counts.attrs["target"] = source_addr  # a NeXus API convention for links
    nxdata[target_addr] = f[source_addr]  # hard link
    # nxdata._id.link(source_addr, target_addr, h5py.h5g.LINK_HARD)



"""This module Generate ELN in a hierarchical format according to NEXUS definition."""
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import click
from pynxtools.eln_mapper.eln import generate_eln
from pynxtools.eln_mapper.scheme_eln import generate_scheme_eln


@click.command()
@click.option(
    '--nxdl',
    required=True,
    help="Name of NeXus definition without extension (.nxdl.xml)."
)
@click.option(
    '--skip-top-levels',
    default=1,
    required=False,
    type=int,
    show_default=True,
    help=("To skip the level of parent hierarchy level. E.g. for default 1 the part"
          "Entry[ENTRY] from /Entry[ENTRY]/Instrument[INSTRUMENT]/... will be skiped.")
)
@click.option(
    '--output-file',
    required=False,
    default='eln_data',
    help=('Name of file that is neede to generated output file.')
)
@click.option(
    '--eln-type',
    required=True,
    type=click.Choice(['eln', 'scheme_eln'], case_sensitive=False),
    default='eln'
)
def get_eln(nxdl: str,
            skip_top_levels: int,
            output_file: str,
            eln_type: str):
    """To generate ELN in yaml file format.

    Parameters
    ----------

        nxdl : str
            Name of NeXus definition e.g. NXmpes
        skip_top_levels : int
            To skip hierarchical levels
        output_file : str
            Name of the output file.
    """
    eln_type = eln_type.lower()
    if eln_type == 'eln':
        generate_eln(nxdl, output_file, skip_top_levels)
    elif eln_type == 'scheme_eln':
        generate_scheme_eln(nxdl, eln_file_name=output_file)


if __name__ == "__main__":
    get_eln().parse()  # pylint: disable=no-value-for-parameter



import pytest

from ..docs import AnchorRegistry
from ..docs import NXClassDocGenerator
from ..docs import XSDDocGenerator
from ..docs import nxdl_indices
from ..globals.directories import get_xsd_file
from ..nxdl import iter_definitions


@pytest.fixture(scope="module")
def doc_generator():
    return NXClassDocGenerator()


@pytest.fixture(scope="module")
def anchor_registry():
    return AnchorRegistry()


@pytest.fixture(scope="module")
def anchor_registry_write(tmpdir_factory):
    tmpdir = tmpdir_factory.mktemp("registry")
    reg = AnchorRegistry(output_path=tmpdir)
    yield reg
    reg.write()


@pytest.mark.parametrize("nxdl_file", list(iter_definitions()))
def test_nxdl_generate_doc(nxdl_file, doc_generator):
    assert doc_generator(nxdl_file)


@pytest.mark.parametrize("nxdl_file", list(iter_definitions()))
def test_nxdl_anchor_list(nxdl_file, doc_generator, anchor_registry):
    assert doc_generator(nxdl_file, anchor_registry=anchor_registry)


@pytest.mark.parametrize("nxdl_file", list(iter_definitions()))
def test_nxdl_anchor_write_list(nxdl_file, doc_generator, anchor_registry_write):
    assert doc_generator(nxdl_file, anchor_registry=anchor_registry_write)


def test_nxdl_indices():
    sections = nxdl_indices()
    expected = {"base_classes", "applications", "contributed_definitions"}
    assert set(sections) == expected


def test_xsd_generate_doc():
    generator = XSDDocGenerator()
    assert generator(get_xsd_file())



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

import sys, os, datetime


# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# add the abs path to the custom extension for collecting the contributor variables from the rst files
sys.path.insert(0, os.path.abspath('../../../dev_tools/ext'))

# -- Project information -----------------------------------------------------

project = 'NeXus-FAIRmat'
author = 'The FAIRmat collaboration'
copyright = u'2022-{}, {}'.format(datetime.datetime.now().year, author)
description = u'Proposal of NeXus expansion for FAIRmat data'

# The full version, including alpha/beta/rc tags
version = u'unknown NXDL version'
release = u'unknown NXDL release'
nxdl_version = open('../../NXDL_VERSION').read().strip()
if nxdl_version is not None:
    version = nxdl_version.split('.')[0]
    release = nxdl_version


# -- General configuration ---------------------------------------------------

# https://github.com/nexusformat/definitions/issues/659#issuecomment-577438319
needs_sphinx = '2.5'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx_toolbox.collapse',
    'sphinx.ext.mathjax',
    'sphinx.ext.ifconfig',
    'sphinx.ext.viewcode',
    'sphinx.ext.githubpages',
    'sphinx_comments',
    'sphinx.ext.todo',
    'sphinx_tabs.tabs',
    'contrib_ext'
]


# Show `.. todo` directives in the output
# todo_include_todos = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'alabaster'
# html_theme = 'sphinxdoc'
# html_theme = 'sphinx_rtd_theme'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# Add extra files
html_extra_path = ['CNAME']
html_logo = 'img/FAIRmat_new.png'

if html_theme== 'sphinx_rtd_theme':
    html_theme_options = {
        'logo_only': False,
        'collapse_navigation': True,
        'sticky_navigation': True,
        'navigation_depth': 4,
        'includehidden': True,
        'titles_only': False
    }
    def setup(app):
        app.add_css_file('to_rtd.css')
elif html_theme== 'alabaster': # Alabaster allows a very high degree of control form Sphinx conf.py
    html_sidebars = {
    '**': [
        'about.html',
        'navigation.html',
        'localtoc.html',
        'relations.html',
        'sourcelink.html',
        'searchbox.html',
        'google_search.html'
        ],
    }
    html_theme_options = {
        'body_text_align': 'justify',
        'logo_name': True,
        'github_button': True,
        'github_type': 'watch',
        'github_repo': 'nexus_definitions/tree/fairmat',
        'github_user': 'FAIRmat-Experimental',
        'github_count': 'false', # We don't get the cute counter baloon if we want to point to the branch
        'sidebar_width': '300px',
        'body_max_width': 'auto',
        'page_width': 'auto',
        'font_size': '12pt',
        'font_family': 'Arial',
        'description': 'Proposal of NeXus expansion for FAIRmat data.',
        'show_powered_by': True,
        'sidebar_header': '#ffffff',
        'sidebar_hr': '#ffffff',
        'sidebar_link': '#ffffff',
        'sidebar_list': '#ffffff',
        'sidebar_link_underscore': '#ffffff',
        'sidebar_text': '#ffffff'
    }
    def setup(app):
        app.add_css_file('to_alabaster.css')
else:
    html_sidebars = {
    '**': [
        'localtoc.html',
        'relations.html',
        'sourcelink.html',
        'searchbox.html',
        'google_search.html'
        ],
    }

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_favicon = "../../../common/favicon.ico"

# Output file base name for HTML help builder.
htmlhelp_basename = 'NeXusManualdoc'

comments_config = {
    "hypothesis": True
}

# -- Options for Latex output -------------------------------------------------
latex_elements = {
    'maxlistdepth': 25, # some application definitions are deeply nested
    'preamble': r'''
    \usepackage{amsbsy}
    \listfiles
    \DeclareUnicodeCharacter{1F517}{X}
    \DeclareUnicodeCharacter{2906}{<=}'''
}



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Helper functions for reading sections from Lake Shore files"""
from typing import Tuple, Union, Dict, Any
import re
from datetime import datetime
import numpy as np
import pandas as pd
import pytz


def is_section(expr: str) -> bool:
    """Checks whether an expression follows the form of a section
    i.e. is of the form [section]

    Args:
        expr (str): The current expression to check

    Returns:
        bool: Returns true if the expr is of the form of a section
    """
    return bool(re.search(r"^\[.+\]$", expr))


def is_measurement(expr):
    """Checks whether an expression follows the form of a measurement indicator
    i.e. is of the form <measurement>

    Args:
        expr (str): The current expression to check

    Returns:
        bool: Returns true if the expr is of the form of a measurement indicator
    """
    return bool(re.search(r"^\<.+\>$", expr))


def is_key(expr: str) -> bool:
    """Checks whether an expression follows the form of a key value pair
    i.e. is of the form key: value or key = value

    Args:
        expr (str): The current expression to check

    Returns:
        bool: Returns true if the expr is of the form of a key value pair
    """
    return bool(re.search(r"^.+\s*[:|=]\s*.+$", expr))


def is_meas_header(expr: str) -> bool:
    """Checks whether an expression follows the form of a measurement header,
    i.e. starts with: Word [Unit]

    Args:
        expr (str): The current expression to check

    Returns:
        bool: Returns true if the expr is of the form of a measurement header
    """
    return bool(re.search(r"^[^\]]+\[[^\]]+\]", expr))


def is_value_with_unit(expr: str) -> bool:
    """Checks whether an expression is a value with a unit,
    i.e. is of the form value [unit].

    Args:
        expr (str): The expression to check

    Returns:
        bool: Returns true if the expr is a value with unit
    """
    return bool(re.search(r"^.+\s\[.+\]$", expr))


def is_integer(expr: str) -> bool:
    """Checks whether an expression is an integer number,
    i.e. 3, +3 or -3. Also supports numbers in the form of 003.

    Args:
        expr (str): The expression to check

    Returns:
        bool: Returns true if the expr is an integer number
    """
    return bool(re.search(r"^[+-]?\d+$", expr))


def is_number(expr: str) -> bool:
    """Checks whether an expression is a number,
    i.e. is of the form 0.3, 3, 1e-3, 1E5 etc.

    Args:
        expr (str): The expression to check

    Returns:
        bool: Returns true if the expr is a number
    """
    return bool(
        re.search(r"^[+-]?(\d+([.]\d*)?([eE][+-]?\d+)?|[.]\d+([eE][+-]?\d+)?)$", expr)
    )


def is_boolean(expr: str) -> bool:
    """Checks whether an expression is a boolean,
    i.e. is equal to True or False (upper or lower case).

    Args:
        expr (str): The expression to check.

    Returns:
        bool: Returns true if the expr is a boolean
    """
    return bool(re.search(r"True|False|true|false|On|Off|Yes|No", expr))


def to_bool(expr: str) -> bool:
    """Converts boolean representations in strings to python booleans.

    Args:
        expr (str): The string to convert to boolean.

    Returns:
        bool: The boolean value.
    """
    replacements = {
        'On': True,
        'Off': False,
        'Yes': True,
        'No': False,
        'True': True,
        'False': False,
        'true': True,
        'false': False,
    }

    return replacements.get(expr)


def split_str_with_unit(expr: str, lower: bool = True) -> Tuple[str, str]:
    """Splits an expression into a string and a unit.
    The input expression should be of the form value [unit] as
    is checked with is_value_with_unit function.

    Args:
        expr (str): The expression to split
        lower (bool, optional):
            If True the value is converted to lower case. Defaults to True.

    Returns:
        Tuple[str, str]: A tuple of a value unit pair.
    """
    value = re.split(r"\s+\[.+\]", expr)[0]
    unit = re.search(r"(?<=\[).+?(?=\])", expr)[0]

    if lower:
        return value.lower(), unit
    return value, unit


def split_value_with_unit(expr: str) -> Tuple[Union[float, str], str]:
    """Splits an expression into a string or float and a unit.
    The input expression should be of the form value [unit] as
    is checked with is_value_with_unit function.
    The value is automatically converted to a float if it is a number.

    Args:
        expr (str): The expression to split

    Returns:
        Tuple[Union[float, str], str]: A tuple of a value unit pair.
    """
    value, unit = split_str_with_unit(expr, False)

    if is_number(value):
        return float(value), unit

    return value, unit


def clean(unit: str) -> str:
    """Cleans an unit string, e.g. converts `VS` to `volt * seconds`.
    If the unit is not in the conversion dict the input string is
    returned without modification.

    Args:
        unit (str): The dirty unit string.

    Returns:
        str: The cleaned unit string.
    """
    conversions = {
        'VS': "volt * second",
        'Sec': "s",
        '²': "^2",
        '³': "^3",
        'ohm cm': "ohm * cm",
    }

    for old, new in conversions.items():
        unit = unit.replace(old, new)

    return unit


def get_unique_dkey(dic: dict, dkey: str) -> str:
    """Checks whether a data key is already contained in a dictionary
    and returns a unique key if it is not.

    Args:
        dic (dict): The dictionary to check for keys
        dkey (str): The data key which shall be written.

    Returns:
        str: A unique data key. If a key already exists it is appended with a number
    """
    suffix = 0
    while f"{dkey}{suffix}" in dic:
        suffix += 1

    return f"{dkey}{suffix}"


def pandas_df_to_template(prefix: str, data: pd.DataFrame) -> Dict[str, Any]:
    """Converts a dataframe to a NXdata entry template.

    Args:
        prefix (str): The path prefix to write the data into. Without a trailing slash.
        df (pd.DataFrame): The dataframe which should be converted.

    Returns:
        Dict[str, Any]: The dict containing the data and metainfo.
    """
    if prefix.endswith('/'):
        prefix = prefix[:-1]

    template: Dict[str, Any] = {}
    template[f'{prefix}/@NX_class'] = 'NXdata'

    def write_data(header: str, attr: str, data: np.ndarray) -> None:
        if header is None:
            print('Warning: Trying to write dataframe without a header. Skipping.')
            return

        if is_value_with_unit(header):
            name, unit = split_str_with_unit(header)
            template[f'{prefix}/{name}/@units'] = clean(unit)
        else:
            name = header.lower()

        if attr == '@auxiliary_signals':
            if f'{prefix}/{attr}' in template:
                template[f'{prefix}/{attr}'].append(name)
            else:
                template[f'{prefix}/{attr}'] = [name]
        else:
            template[f'{prefix}/{attr}'] = name
        template[f'{prefix}/{name}'] = data

    if data.index.name is None:
        data = data.set_index(data.columns[0])

    # Drop last line if it has an errornous zero temperature
    if data.index.values[-1] == 0:
        data = data.iloc[:-1]

    write_data(data.index.name, '@axes', data.index.values)
    write_data(data.columns[0], '@signal', data.iloc[:, 0].values)

    for column in data.columns[1:]:
        write_data(column, '@auxiliary_signals', data[column].values)

    return template


def convert_date(datestr: str, timezone: str = "Europe/Berlin") -> str:
    """Converts a hall date formated string to isoformat string.

    Args:
        datestr (str): The hall date string
        timezone (str): The timezone of the hall date string. Defaults to "Europe/Berlin"

    Returns:
        str: The iso formatted string.
    """

    try:
        return (
            datetime
            .strptime(datestr, r'%m/%d/%y %H%M%S')
            .astimezone(pytz.timezone(timezone))
            .isoformat()
        )
    except ValueError:
        print("Warning: datestring does not conform to date format. Skipping.")
        return datestr



import shutil
from pathlib import Path
from typing import List
from typing import Tuple

from ..globals import directories


def copyfile(from_path: Path, to_path: Path) -> None:
    print("copy", from_path)
    print("  ->", to_path)
    shutil.copyfile(from_path, to_path)


def copydir(from_path: Path, to_path: Path) -> None:
    print("copy", from_path)
    print("  ->", to_path)
    shutil.copytree(from_path, to_path, dirs_exist_ok=True)


def copy_files(files: List[Tuple[str, str, bool]]) -> None:
    source_root = directories.get_source_root()
    build_root = directories.get_build_root()
    for from_subname, to_subname, overwrite in files:
        to_path = build_root / to_subname
        if overwrite or not to_path.exists():
            from_path = source_root / from_subname
            copyfile(from_path, to_path)
        else:
            print("already exists", to_path)



"""
Read the the NeXus NXDL types specification and find
all the valid types of units.  Write a restructured
text (.rst) document for use in the NeXus manual in
the NXDL chapter.
"""


from pathlib import Path
from typing import List

import lxml.etree

from ..globals.nxdl import XSD_NAMESPACE
from ..utils.types import PathLike


def generate_xsd_units_doc(
    xsd_file: PathLike, nodeMatchString: str, section: str
) -> List[str]:
    xsd_file = str(Path(xsd_file))
    tree = lxml.etree.parse(xsd_file)

    rst_lines = [f".. auto-generated by {__name__} -- DO NOT EDIT\n"]
    rst_lines.append("\n")

    rst_lines.append(f".. nodeMatchString : {nodeMatchString}\n")
    rst_lines.append("\n")

    db = {}
    ns = {"xs": XSD_NAMESPACE}
    node_list = tree.xpath("//xs:simpleType", namespaces=ns)

    # get the names of all the types of units
    members = []
    for node in node_list:
        if node.get("name") == nodeMatchString:
            union = node.xpath("xs:union", namespaces=ns)
            members = union[0].get("memberTypes", "").split()

    # get the definition of each type of units
    for node in node_list:
        node_name = node.get("name")
        if node_name is None:
            continue
        if "nxdl:" + node_name in members:
            words = node.xpath("xs:annotation/xs:documentation", namespaces=ns)[0]
            examples = []
            for example in words.iterchildren():
                nm = example.attrib.get("name")
                if nm is not None and nm == "example":
                    examples.append("``" + example.text + "``")
            a = words.text
            if len(examples) > 0:
                a = " ".join(a.split()) + ",\n\texample(s): " + " | ".join(examples)
            db[node_name] = a

            # for item in node.xpath("xs:restriction//xs:enumeration", namespaces=ns):
            #    key = "%s" % item.get("value")
            #    words = item.xpath("xs:annotation/xs:documentation", namespaces=ns)[0]
            #    db[key] = words.text

    # this list is too long to make this a table in latex
    # for two columns, a Sphinx fieldlist will do just as well
    for key in sorted(db):
        rst_lines.append(f".. index:: ! {key} ({section} type)\n\n")  # index entry
        rst_lines.append(f".. _{key}:\n\n")  # cross-reference point
        rst_lines.append(f":{key}:\n")
        for line in db[key].splitlines():
            rst_lines.append(f"    {line}\n")
        rst_lines.append("\n")
    return rst_lines



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

#
"""An example reader implementation for the DataConverter."""
import os
from typing import Tuple, Any
import json
import numpy as np

from pynxtools.dataconverter.readers.base.reader import BaseReader


class ExampleReader(BaseReader):
    """An example reader implementation for the DataConverter."""

    # pylint: disable=too-few-public-methods

    # Whitelist for the NXDLs that the reader supports and can process
    supported_nxdls = ["NXtest"]

    def read(self,
             template: dict = None,
             file_paths: Tuple[str] = None,
             objects: Tuple[Any] = None) -> dict:
        """Reads data from given file and returns a filled template dictionary"""
        data: dict = {}

        if not file_paths:
            raise IOError("No input files were given to Example Reader.")

        for file_path in file_paths:
            file_extension = file_path[file_path.rindex("."):]
            with open(file_path, "r", encoding="utf-8") as input_file:
                if file_extension == ".json":
                    data = json.loads(input_file.read())

        for k in template.keys():
            # The entries in the template dict should correspond with what the dataconverter
            # outputs with --generate-template for a provided NXDL file
            if k.startswith("/ENTRY[entry]/required_group") \
               or k == "/ENTRY[entry]/optional_parent/req_group_in_opt_group":
                continue

            field_name = k[k.rfind("/") + 1:]
            if field_name != "@units":
                template[k] = data[field_name]
                if f"{field_name}_units" in data.keys() and f"{k}/@units" in template.keys():
                    template[f"{k}/@units"] = data[f"{field_name}_units"]

        template["required"]["/ENTRY[entry]/optional_parent/required_child"] = 1
        template["optional"][("/ENTRY[entry]/optional_parent/"
                              "req_group_in_opt_group/DATA[data]")] = [0, 1]

        # Add non template key
        template["/ENTRY[entry]/does/not/exist"] = "None"
        template["/ENTRY[entry]/required_group/description"] = "A test description"
        template["/ENTRY[entry]/required_group2/description"] = "A test description"
        template["/ENTRY[entry]/program_name"] = "None"

        # internal links
        template["/ENTRY[entry]/test_link/internal_link"] = {"link":
                                                             "/entry/NXODD_name/posint_value"}

        # external links
        template[("/ENTRY"
                  "[entry]/test_link"
                  "/external_link")] = {"link":
                                        f"{os.path.dirname(__file__)}/../../../../tests/"
                                        f"data/dataconverter/readers/mpes/"
                                        f"xarray_saved_small_calibration.h5:/axes/ax3"
                                        }

        # virtual datasets concatenation
        my_path = str(f"{os.path.dirname(__file__)}/../../../../tests/"
                      f"data/dataconverter/readers/mpes")
        my_datasets = {"link":
                       [f"{my_path}/xarray_saved_small_calibration.h5:/axes/ax0",
                        f"{my_path}/xarray_saved_small_calibration.h5:/axes/ax1",
                        f"{my_path}/xarray_saved_small_calibration.h5:/axes/ax2"
                        ]
                       }
        template["/ENTRY[entry]/test_virtual_dataset/concatenate_datasets"] = my_datasets

        # virtual datasets slicing
        my_path = str(f"{os.path.dirname(__file__)}/../../../../tests/"
                      f"data/dataconverter/readers/mpes")
        template[("/ENTRY[entry]"
                  "/test_virtual"
                  "_dataset/sliced"
                  "_dataset")] = {"link":
                                  (f"{my_path}/xarray_saved_small_"
                                   "calibration.h5:/binned/BinnedData"),
                                  "shape": np.index_exp[:, 1, :, :]
                                  }
        template[("/ENTRY[entry]"
                  "/test_virtual"
                  "_dataset/slic"
                  "ed_dataset2")] = {"link":
                                     (f"{my_path}/xarray_saved_small"
                                      "_calibration.h5:/binned/BinnedData"),
                                     "shape": np.index_exp[:, :, :, 1]
                                     }
        template[("/ENTRY[entry]"
                  "/test_virtual"
                  "_dataset/slic"
                  "ed_dataset3")] = {"link":
                                     (f"{my_path}/xarray_saved_small"
                                      "_calibration.h5:/binned/BinnedData"),
                                     "shape": np.index_exp[:, :, :, 2:4]
                                     }

        # compression
        my_compression_dict = {"compress": "string not to be compressed"}
        template["/ENTRY[entry]/test_compression/not_to_compress"] = my_compression_dict
        my_compression_dict2 = {"compress": np.array([1, 2, 3, 4])}
        template["/ENTRY[entry]/test_compression/compressed_data"] = my_compression_dict2

        # sh = h5py.File(file_names_to_concatenate[0], 'r')[entry_key].shape
        # layout = h5py.VirtualLayout(shape=(len(file_names_to_concatenate),) + sh,
        #                             dtype=np.float64)

        return template


# This has to be set to allow the convert script to use this reader. Set it to "MyDataReader".
READER = ExampleReader



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Utility for analyzing swift data/metadata of display_items to identify NeXus concepts."""

# pylint: disable=no-member

import flatdict as fd


metadata_constraints = {"type": str,
                        "uuid": str,
                        "created": str,
                        "data_shape": list,
                        "data_dtype": str,
                        "is_sequence": bool,
                        "dimensional_calibrations": list,
                        "data_modified": str,
                        "timezone": str,
                        "timezone_offset": str,
                        "metadata/hardware_source/hardware_source_id": str,
                        "version": int,
                        "modified": str}

nexus_concept_dict = {"ITULL": "NxImageSetRealSpace",
                      "IFLL": "NxImageSetRealSpace",
                      "IFL": None,
                      "ITUL": None,
                      "STUUE": "NxSpectrumSetEelsOmegaQ",
                      "STULLE": "NxSpectrumSetEels",
                      "STULLUE": "NxSpectrumSetOmegaQ",
                      "SFLLUE": "NxSpectrumSetOmegaQ",
                      "SFLLE": "NxSpectrumSetEels",
                      "SFUE": "NxSpectrumSetEelsOmegaQ",
                      "RFAA": "NxImageAngSpace",
                      "RTUAA": "NxImageAngSpace"}


def check_existence_of_required_fields(dct: dict, constraint_dct: dict) -> bool:
    """Checks if given dictionary has fields with values which match constraints."""
    flat_dct = fd.FlatDict(dct, delimiter='/')
    for keyword, dtyp in constraint_dct.items():
        if keyword not in flat_dct.keys():
            print(f"-->{keyword} not keyword")
            return False
        if not isinstance(flat_dct[keyword], dtyp):
            print(f"-->{keyword} not instance")
            return False
    return True


def identify_nexus_concept_key(dct: dict) -> str:
    """Identifies best candidate to map data/metadata on a NeXus concept."""
    # ##MK::imporve that we work ideally always with the flattened dictionary
    nexus_concept_key = "UNKNOWN"
    if check_existence_of_required_fields(dct, metadata_constraints) is False:
        return nexus_concept_key
    lst_unit_catg = []
    for axis_dict in dct["dimensional_calibrations"]:  # inspect axes in sequence
        if isinstance(axis_dict, dict):
            if set(axis_dict.keys()) == set(["offset", "scale", "units"]):
                unit_arg = axis_dict["units"].lower()
                if unit_arg == "":
                    lst_unit_catg.append("U")
                elif unit_arg in ["nm"]:  # replace by pint to pick up on any length
                    lst_unit_catg.append("L")
                elif unit_arg in ["ev"]:  # replace by pint to pick up on any enery
                    lst_unit_catg.append("E")
                elif unit_arg in ["rad"]:  # replace by pint to pick up on angle unit
                    lst_unit_catg.append("A")
                else:
                    return nexus_concept_key
        set_unit_catg = set(lst_unit_catg)

        if "A" in set_unit_catg:
            nexus_concept_key \
                = f"R{str(dct['is_sequence']).upper()[0:1]}{''.join(lst_unit_catg)}"
        elif "E" in set_unit_catg:
            nexus_concept_key \
                = f"S{str(dct['is_sequence']).upper()[0:1]}{''.join(lst_unit_catg)}"
        elif "E" not in set_unit_catg:
            nexus_concept_key \
                = f"I{str(dct['is_sequence']).upper()[0:1]}{''.join(lst_unit_catg)}"
        else:
            return nexus_concept_key
    return nexus_concept_key
